from ... import platforms
from .. import state
from ..x86_registers import X86MMRRegister
from . import cpu


class AMD64(cpu.CPU):
    """Generic AMD64 CPU state model.

    Specific implementations support different vector extensions.
    Because of how smallworld works, an emulator can only support
    platforms if it supports all base registers.
    Since the AVX extensions keep adding registers under
    the old ones, we need new platforms.
    """

    def __init__(self):
        super().__init__()
        # *** General Purpose Registers ***
        self.rax = state.Register("rax", 8)
        self.add(self.rax)
        self.eax = state.RegisterAlias("eax", self.rax, 4, 0)
        self.add(self.eax)
        self.ax = state.RegisterAlias("ax", self.rax, 2, 0)
        self.add(self.ax)
        self.al = state.RegisterAlias("al", self.rax, 1, 0)
        self.add(self.al)
        self.ah = state.RegisterAlias("ah", self.rax, 1, 1)
        self.add(self.ah)

        self.rbx = state.Register("rbx", 8)
        self.add(self.rbx)
        self.ebx = state.RegisterAlias("ebx", self.rbx, 4, 0)
        self.add(self.ebx)
        self.bx = state.RegisterAlias("bx", self.rbx, 2, 0)
        self.add(self.bx)
        self.bl = state.RegisterAlias("bl", self.rbx, 1, 0)
        self.add(self.bl)
        self.bh = state.RegisterAlias("bh", self.rbx, 1, 1)
        self.add(self.bh)

        self.rcx = state.Register("rcx", 8)
        self.add(self.rcx)
        self.ecx = state.RegisterAlias("ecx", self.rcx, 4, 0)
        self.add(self.ecx)
        self.cx = state.RegisterAlias("cx", self.rcx, 2, 0)
        self.add(self.cx)
        self.cl = state.RegisterAlias("cl", self.rcx, 1, 0)
        self.add(self.cl)
        self.ch = state.RegisterAlias("ch", self.rcx, 1, 1)
        self.add(self.ch)

        self.rdx = state.Register("rdx", 8)
        self.add(self.rdx)
        self.edx = state.RegisterAlias("edx", self.rdx, 4, 0)
        self.add(self.edx)
        self.dx = state.RegisterAlias("dx", self.rdx, 2, 0)
        self.add(self.dx)
        self.dl = state.RegisterAlias("dl", self.rdx, 1, 0)
        self.add(self.dl)
        self.dh = state.RegisterAlias("dh", self.rdx, 1, 1)
        self.add(self.dh)

        self.r8 = state.Register("r8", 8)
        self.add(self.r8)
        self.r8d = state.RegisterAlias("r8d", self.r8, 4, 0)
        self.add(self.r8d)
        self.r8w = state.RegisterAlias("r8w", self.r8, 2, 0)
        self.add(self.r8w)
        self.r8b = state.RegisterAlias("r8b", self.r8, 1, 0)
        self.add(self.r8b)

        self.r9 = state.Register("r9", 8)
        self.add(self.r9)
        self.r9d = state.RegisterAlias("r9d", self.r9, 4, 0)
        self.add(self.r9d)
        self.r9w = state.RegisterAlias("r9w", self.r9, 2, 0)
        self.add(self.r9w)
        self.r9b = state.RegisterAlias("r9b", self.r9, 1, 0)
        self.add(self.r9b)

        self.r10 = state.Register("r10", 8)
        self.add(self.r10)
        self.r10d = state.RegisterAlias("r10d", self.r10, 4, 0)
        self.add(self.r10d)
        self.r10w = state.RegisterAlias("r10w", self.r10, 2, 0)
        self.add(self.r10w)
        self.r10b = state.RegisterAlias("r10b", self.r10, 1, 0)
        self.add(self.r10b)

        self.r11 = state.Register("r11", 8)
        self.add(self.r11)
        self.r11d = state.RegisterAlias("r11d", self.r11, 4, 0)
        self.add(self.r11d)
        self.r11w = state.RegisterAlias("r11w", self.r11, 2, 0)
        self.add(self.r11w)
        self.r11b = state.RegisterAlias("r11b", self.r11, 1, 0)
        self.add(self.r11b)

        self.r12 = state.Register("r12", 8)
        self.add(self.r12)
        self.r12d = state.RegisterAlias("r12d", self.r12, 4, 0)
        self.add(self.r12d)
        self.r12w = state.RegisterAlias("r12w", self.r12, 2, 0)
        self.add(self.r12w)
        self.r12b = state.RegisterAlias("r12b", self.r12, 1, 0)
        self.add(self.r12b)

        self.r13 = state.Register("r13", 8)
        self.add(self.r13)
        self.r13d = state.RegisterAlias("r13d", self.r13, 4, 0)
        self.add(self.r13d)
        self.r13w = state.RegisterAlias("r13w", self.r13, 2, 0)
        self.add(self.r13w)
        self.r13b = state.RegisterAlias("r13b", self.r13, 1, 0)
        self.add(self.r13b)

        self.r14 = state.Register("r14", 8)
        self.add(self.r14)
        self.r14d = state.RegisterAlias("r14d", self.r14, 4, 0)
        self.add(self.r14d)
        self.r14w = state.RegisterAlias("r14w", self.r14, 2, 0)
        self.add(self.r14w)
        self.r14b = state.RegisterAlias("r14b", self.r14, 1, 0)
        self.add(self.r14b)

        self.r15 = state.Register("r15", 8)
        self.add(self.r15)
        self.r15d = state.RegisterAlias("r15d", self.r15, 4, 0)
        self.add(self.r15d)
        self.r15w = state.RegisterAlias("r15w", self.r15, 2, 0)
        self.add(self.r15w)
        self.r15b = state.RegisterAlias("r15b", self.r15, 1, 0)
        self.add(self.r15b)

        self.rdi = state.Register("rdi", 8)
        self.add(self.rdi)
        self.edi = state.RegisterAlias("edi", self.rdi, 4, 0)
        self.add(self.edi)
        self.di = state.RegisterAlias("di", self.rdi, 2, 0)
        self.add(self.di)
        self.dil = state.RegisterAlias("dil", self.rdi, 1, 0)
        self.add(self.dil)

        self.rsi = state.Register("rsi", 8)
        self.add(self.rsi)
        self.esi = state.RegisterAlias("esi", self.rsi, 4, 0)
        self.add(self.esi)
        self.si = state.RegisterAlias("si", self.rsi, 2, 0)
        self.add(self.si)
        self.sil = state.RegisterAlias("sil", self.rsi, 1, 0)
        self.add(self.sil)

        self.rsp = state.Register("rsp", 8)
        self.add(self.rsp)
        self.esp = state.RegisterAlias("esp", self.rsp, 4, 0)
        self.add(self.esp)
        self.sp = state.RegisterAlias("sp", self.rsp, 2, 0)
        self.add(self.sp)
        self.spl = state.RegisterAlias("spl", self.rsp, 1, 0)
        self.add(self.spl)

        self.rbp = state.Register("rbp", 8)
        self.add(self.rbp)
        self.ebp = state.RegisterAlias("ebp", self.rbp, 4, 0)
        self.add(self.ebp)
        self.bp = state.RegisterAlias("bp", self.rbp, 2, 0)
        self.add(self.bp)
        self.bpl = state.RegisterAlias("bpl", self.rbp, 1, 0)
        self.add(self.bpl)

        # *** Instruction Pointer ***
        self.rip = state.Register("rip", 8)
        self.add(self.rip)
        self.eip = state.RegisterAlias("eip", self.rip, 4, 0)
        self.add(self.eip)
        self.ip = state.RegisterAlias("ip", self.rip, 2, 0)
        self.add(self.ip)

        self.pc = state.RegisterAlias("pc", self.rip, 8, 0)
        self.add(self.pc)

        # *** Flags register ***
        self.rflags = state.Register("rflags", 8)
        self.add(self.rflags)
        self.eflags = state.RegisterAlias("eflags", self.rflags, 4, 0)
        self.add(self.eflags)
        self.flags = state.RegisterAlias("flags", self.rflags, 2, 0)
        self.add(self.flags)

        # *** Segment Registers ***
        # NOTE: These are actually 16 bits
        # However, their representation in different emulators gets weird.
        self.cs = state.Register("cs", 8)
        self.add(self.cs)
        self.ds = state.Register("ds", 8)
        self.add(self.ds)
        self.es = state.Register("es", 8)
        self.add(self.es)
        self.fs = state.Register("fs", 8)
        self.add(self.fs)
        self.gs = state.Register("gs", 8)
        self.add(self.gs)
        self.ss = state.Register("ss", 8)
        self.add(self.ss)

        # *** Control Registers ***
        self.cr0 = state.Register("cr0", 8)
        self.add(self.cr0)
        self.cr1 = state.Register("cr1", 8)
        self.add(self.cr1)
        self.cr2 = state.Register("cr2", 8)
        self.add(self.cr2)
        self.cr3 = state.Register("cr3", 8)
        self.add(self.cr3)
        self.cr4 = state.Register("cr4", 8)
        self.add(self.cr4)
        self.cr8 = state.Register("cr8", 8)
        self.add(self.cr8)

        # *** Debug Registers ***
        self.dr0 = state.Register("dr0", 8)
        self.add(self.dr0)
        self.dr1 = state.Register("dr1", 8)
        self.add(self.dr1)
        self.dr2 = state.Register("dr2", 8)
        self.add(self.dr2)
        self.dr3 = state.Register("dr3", 8)
        self.add(self.dr3)
        self.dr6 = state.Register("dr6", 8)
        self.add(self.dr6)
        self.dr7 = state.Register("dr7", 8)
        self.add(self.dr7)
        self.dr8 = state.Register("dr8", 8)
        self.add(self.dr8)
        self.dr9 = state.Register("dr9", 8)
        self.add(self.dr9)
        self.dr10 = state.Register("dr10", 8)
        self.add(self.dr10)
        self.dr11 = state.Register("dr11", 8)
        self.add(self.dr11)
        self.dr12 = state.Register("dr12", 8)
        self.add(self.dr12)
        self.dr13 = state.Register("dr13", 8)
        self.add(self.dr13)
        self.dr14 = state.Register("dr14", 8)
        self.add(self.dr14)
        self.dr15 = state.Register("dr15", 8)
        self.add(self.dr15)

        # *** Descriptor Table Registers ***
        self.gdtr = X86MMRRegister("gdtr", 10)
        self.add(self.gdtr)
        self.idtr = X86MMRRegister("idtr", 10)
        self.add(self.idtr)
        self.ldtr = X86MMRRegister("ldtr", 10)
        self.add(self.ldtr)

        # *** Task Register ***
        self.tr = X86MMRRegister("tr", 2)
        self.add(self.tr)

        # *** x87 registers ***
        self.fpr0 = state.Register("fpr0", 10)
        self.add(self.fpr0)
        self.fpr1 = state.Register("fpr1", 10)
        self.add(self.fpr1)
        self.fpr2 = state.Register("fpr2", 10)
        self.add(self.fpr2)
        self.fpr3 = state.Register("fpr3", 10)
        self.add(self.fpr3)
        self.fpr4 = state.Register("fpr4", 10)
        self.add(self.fpr4)
        self.fpr5 = state.Register("fpr5", 10)
        self.add(self.fpr5)
        self.fpr6 = state.Register("fpr6", 10)
        self.add(self.fpr6)
        self.fpr7 = state.Register("fpr7", 10)
        self.add(self.fpr7)

        # x87 Control Register
        self.fctrl = state.Register("fctrl", 2)
        self.add(self.fctrl)
        # x87 Status Register
        self.fstat = state.Register("fstat", 2)
        self.add(self.fstat)
        # x87 Tag Register
        self.ftag = state.Register("ftag", 2)
        self.add(self.ftag)
        # x87 Last Instruction Register
        self.fip = state.Register("fip", 8)
        self.add(self.fip)
        # x87 Last Operand Pointer
        self.fdp = state.Register("fdp", 8)
        self.add(self.fdp)
        # x87 Last Opcode
        self.fop = state.Register("fop", 2)
        self.add(self.fop)

        # NOTE: Docs disagree on the format of fip and fdp.
        # One source describes them as 48-bit offset-plus-segment,
        # the other describes them as 64-bit.
        # There may also be separate segment registers.
        # If you care about the x87 debug info, please feel free to update.

        # *** MMX Registers ***
        # NOTE: The MMX registers are aliases for the low 8 bytes of the x87 registers.
        # The two subsystems cannot be used simultaneously.
        self.mm0 = state.RegisterAlias("mm0", self.fpr0, 8, 0)
        self.add(self.mm0)
        self.mm1 = state.RegisterAlias("mm1", self.fpr1, 8, 0)
        self.add(self.mm1)
        self.mm2 = state.RegisterAlias("mm2", self.fpr2, 8, 0)
        self.add(self.mm2)
        self.mm3 = state.RegisterAlias("mm3", self.fpr3, 8, 0)
        self.add(self.mm3)
        self.mm4 = state.RegisterAlias("mm4", self.fpr4, 8, 0)
        self.add(self.mm4)
        self.mm5 = state.RegisterAlias("mm5", self.fpr5, 8, 0)
        self.add(self.mm5)
        self.mm6 = state.RegisterAlias("mm6", self.fpr6, 8, 0)
        self.add(self.mm6)
        self.mm7 = state.RegisterAlias("mm7", self.fpr7, 8, 0)
        self.add(self.mm7)


class AMD64AVX2(AMD64):
    """AMD64 CPU supporting up to AVX2

    This is our default, since all emulators support up to AVX2,
    and 99.9% of our users won't use the vector extensions.
    """

    platform = platforms.Platform(
        platforms.Architecture.X86_64, platforms.Byteorder.LITTLE
    )

    def __init__(self):
        super().__init__()
        # *** SSE/AVX/AVX2 registers ***
        self.ymm0 = state.Register("ymm0", 32)
        self.add(self.ymm0)
        self.xmm0 = state.RegisterAlias("xmm0", self.ymm0, 16, 0)
        self.add(self.xmm0)

        self.ymm1 = state.Register("ymm1", 32)
        self.add(self.ymm1)
        self.xmm1 = state.RegisterAlias("xmm1", self.ymm1, 16, 0)
        self.add(self.xmm1)

        self.ymm2 = state.Register("ymm2", 32)
        self.add(self.ymm2)
        self.xmm2 = state.RegisterAlias("xmm2", self.ymm2, 16, 0)
        self.add(self.xmm2)

        self.ymm3 = state.Register("ymm3", 32)
        self.add(self.ymm3)
        self.xmm3 = state.RegisterAlias("xmm3", self.ymm3, 16, 0)
        self.add(self.xmm3)

        self.ymm4 = state.Register("ymm4", 32)
        self.add(self.ymm4)
        self.xmm4 = state.RegisterAlias("xmm4", self.ymm4, 16, 0)
        self.add(self.xmm4)

        self.ymm5 = state.Register("ymm5", 32)
        self.add(self.ymm5)
        self.xmm5 = state.RegisterAlias("xmm5", self.ymm5, 16, 0)
        self.add(self.xmm5)

        self.ymm6 = state.Register("ymm6", 32)
        self.add(self.ymm6)
        self.xmm6 = state.RegisterAlias("xmm6", self.ymm6, 16, 0)
        self.add(self.xmm6)

        self.ymm7 = state.Register("ymm7", 32)
        self.add(self.ymm7)
        self.xmm7 = state.RegisterAlias("xmm7", self.ymm7, 16, 0)
        self.add(self.xmm7)

        self.ymm8 = state.Register("ymm8", 32)
        self.add(self.ymm8)
        self.xmm8 = state.RegisterAlias("xmm8", self.ymm8, 16, 0)
        self.add(self.xmm8)

        self.ymm9 = state.Register("ymm9", 32)
        self.add(self.ymm9)
        self.xmm9 = state.RegisterAlias("xmm9", self.ymm9, 16, 0)
        self.add(self.xmm9)

        self.ymm10 = state.Register("ymm10", 32)
        self.add(self.ymm10)
        self.xmm10 = state.RegisterAlias("xmm10", self.ymm10, 16, 0)
        self.add(self.xmm10)

        self.ymm11 = state.Register("ymm11", 32)
        self.add(self.ymm11)
        self.xmm11 = state.RegisterAlias("xmm11", self.ymm11, 16, 0)
        self.add(self.xmm11)

        self.ymm12 = state.Register("ymm12", 32)
        self.add(self.ymm12)
        self.xmm12 = state.RegisterAlias("xmm12", self.ymm12, 16, 0)
        self.add(self.xmm12)

        self.ymm13 = state.Register("ymm13", 32)
        self.add(self.ymm13)
        self.xmm13 = state.RegisterAlias("xmm13", self.ymm13, 16, 0)
        self.add(self.xmm13)

        self.ymm14 = state.Register("ymm14", 32)
        self.add(self.ymm14)
        self.xmm14 = state.RegisterAlias("xmm14", self.ymm14, 16, 0)
        self.add(self.xmm14)

        self.ymm15 = state.Register("ymm15", 32)
        self.add(self.ymm15)
        self.xmm15 = state.RegisterAlias("xmm15", self.ymm15, 16, 0)
        self.add(self.xmm15)


class AMD64AVX512(AMD64):
    """AMD64 CPU supporting up to AVX512"""

    platform = platforms.Platform(
        platforms.Architecture.X86_64_AVX512, platforms.Byteorder.LITTLE
    )

    def __init__(self):
        super().__init__()
        # *** SSE/AVX/AVX2/AVX512 registers ***
        self.zmm0 = state.Register("zmm0", 64)
        self.add(self.zmm0)
        self.ymm0 = state.RegisterAlias("ymm0", self.zmm0, 32, 0)
        self.add(self.ymm0)
        self.xmm0 = state.RegisterAlias("xmm0", self.zmm0, 16, 0)
        self.add(self.xmm0)

        self.zmm1 = state.Register("zmm1", 64)
        self.add(self.zmm1)
        self.ymm1 = state.RegisterAlias("ymm1", self.zmm1, 32, 0)
        self.add(self.ymm1)
        self.xmm1 = state.RegisterAlias("xmm1", self.zmm1, 16, 0)
        self.add(self.xmm1)

        self.zmm2 = state.Register("zmm2", 64)
        self.add(self.zmm2)
        self.ymm2 = state.RegisterAlias("ymm2", self.zmm2, 32, 0)
        self.add(self.ymm2)
        self.xmm2 = state.RegisterAlias("xmm2", self.zmm2, 16, 0)
        self.add(self.xmm2)

        self.zmm3 = state.Register("zmm3", 64)
        self.add(self.zmm3)
        self.ymm3 = state.RegisterAlias("ymm3", self.zmm3, 32, 0)
        self.add(self.ymm3)
        self.xmm3 = state.RegisterAlias("xmm3", self.zmm3, 16, 0)
        self.add(self.xmm3)

        self.zmm4 = state.Register("zmm4", 64)
        self.add(self.zmm4)
        self.ymm4 = state.RegisterAlias("ymm4", self.zmm4, 32, 0)
        self.add(self.ymm4)
        self.xmm4 = state.RegisterAlias("xmm4", self.zmm4, 16, 0)
        self.add(self.xmm4)

        self.zmm5 = state.Register("zmm5", 64)
        self.add(self.zmm5)
        self.ymm5 = state.RegisterAlias("ymm5", self.zmm5, 32, 0)
        self.add(self.ymm5)
        self.xmm5 = state.RegisterAlias("xmm5", self.zmm5, 16, 0)
        self.add(self.xmm5)

        self.zmm6 = state.Register("zmm6", 64)
        self.add(self.zmm6)
        self.ymm6 = state.RegisterAlias("ymm6", self.zmm6, 32, 0)
        self.add(self.ymm6)
        self.xmm6 = state.RegisterAlias("xmm6", self.zmm6, 16, 0)
        self.add(self.xmm6)

        self.zmm7 = state.Register("zmm7", 64)
        self.add(self.zmm7)
        self.ymm7 = state.RegisterAlias("ymm7", self.zmm7, 32, 0)
        self.add(self.ymm7)
        self.xmm7 = state.RegisterAlias("xmm7", self.zmm7, 16, 0)
        self.add(self.xmm7)

        self.zmm8 = state.Register("zmm8", 64)
        self.add(self.zmm8)
        self.ymm8 = state.RegisterAlias("ymm8", self.zmm8, 32, 0)
        self.add(self.ymm8)
        self.xmm8 = state.RegisterAlias("xmm8", self.zmm8, 16, 0)
        self.add(self.xmm8)

        self.zmm9 = state.Register("zmm9", 64)
        self.add(self.zmm9)
        self.ymm9 = state.RegisterAlias("ymm9", self.zmm9, 32, 0)
        self.add(self.ymm9)
        self.xmm9 = state.RegisterAlias("xmm9", self.zmm9, 16, 0)
        self.add(self.xmm9)

        self.zmm10 = state.Register("zmm10", 64)
        self.add(self.zmm10)
        self.ymm10 = state.RegisterAlias("ymm10", self.zmm10, 32, 0)
        self.add(self.ymm10)
        self.xmm10 = state.RegisterAlias("xmm10", self.zmm10, 16, 0)
        self.add(self.xmm10)

        self.zmm11 = state.Register("zmm11", 64)
        self.add(self.zmm11)
        self.ymm11 = state.RegisterAlias("ymm11", self.zmm11, 32, 0)
        self.add(self.ymm11)
        self.xmm11 = state.RegisterAlias("xmm11", self.zmm11, 16, 0)
        self.add(self.xmm11)

        self.zmm12 = state.Register("zmm12", 64)
        self.add(self.zmm12)
        self.ymm12 = state.RegisterAlias("ymm12", self.zmm12, 32, 0)
        self.add(self.ymm12)
        self.xmm12 = state.RegisterAlias("xmm12", self.zmm12, 16, 0)
        self.add(self.xmm12)

        self.zmm13 = state.Register("zmm13", 64)
        self.add(self.zmm13)
        self.ymm13 = state.RegisterAlias("ymm13", self.zmm13, 32, 0)
        self.add(self.ymm13)
        self.xmm13 = state.RegisterAlias("xmm13", self.zmm13, 16, 0)
        self.add(self.xmm13)

        self.zmm14 = state.Register("zmm14", 64)
        self.add(self.zmm14)
        self.ymm14 = state.RegisterAlias("ymm14", self.zmm14, 32, 0)
        self.add(self.ymm14)
        self.xmm14 = state.RegisterAlias("xmm14", self.zmm14, 16, 0)
        self.add(self.xmm14)

        self.zmm15 = state.Register("zmm15", 64)
        self.add(self.zmm15)
        self.ymm15 = state.RegisterAlias("ymm15", self.zmm15, 32, 0)
        self.add(self.ymm15)
        self.xmm15 = state.RegisterAlias("xmm15", self.zmm15, 16, 0)
        self.add(self.xmm15)

        self.zmm16 = state.Register("zmm16", 64)
        self.add(self.zmm16)
        self.ymm16 = state.RegisterAlias("ymm16", self.zmm16, 32, 0)
        self.add(self.ymm16)
        self.xmm16 = state.RegisterAlias("xmm16", self.zmm16, 16, 0)
        self.add(self.xmm16)

        self.zmm17 = state.Register("zmm17", 64)
        self.add(self.zmm17)
        self.ymm17 = state.RegisterAlias("ymm17", self.zmm17, 32, 0)
        self.add(self.ymm17)
        self.xmm17 = state.RegisterAlias("xmm17", self.zmm17, 16, 0)
        self.add(self.xmm17)

        self.zmm18 = state.Register("zmm18", 64)
        self.add(self.zmm18)
        self.ymm18 = state.RegisterAlias("ymm18", self.zmm18, 32, 0)
        self.add(self.ymm18)
        self.xmm18 = state.RegisterAlias("xmm18", self.zmm18, 16, 0)
        self.add(self.xmm18)

        self.zmm19 = state.Register("zmm19", 64)
        self.add(self.zmm19)
        self.ymm19 = state.RegisterAlias("ymm19", self.zmm19, 32, 0)
        self.add(self.ymm19)
        self.xmm19 = state.RegisterAlias("xmm19", self.zmm19, 16, 0)
        self.add(self.xmm19)

        self.zmm20 = state.Register("zmm20", 64)
        self.add(self.zmm20)
        self.ymm20 = state.RegisterAlias("ymm20", self.zmm20, 32, 0)
        self.add(self.ymm20)
        self.xmm20 = state.RegisterAlias("xmm20", self.zmm20, 16, 0)
        self.add(self.xmm20)

        self.zmm21 = state.Register("zmm21", 64)
        self.add(self.zmm21)
        self.ymm21 = state.RegisterAlias("ymm21", self.zmm21, 32, 0)
        self.add(self.ymm21)
        self.xmm21 = state.RegisterAlias("xmm21", self.zmm21, 16, 0)
        self.add(self.xmm21)

        self.zmm22 = state.Register("zmm22", 64)
        self.add(self.zmm22)
        self.ymm22 = state.RegisterAlias("ymm22", self.zmm22, 32, 0)
        self.add(self.ymm22)
        self.xmm22 = state.RegisterAlias("xmm22", self.zmm22, 16, 0)
        self.add(self.xmm22)

        self.zmm23 = state.Register("zmm23", 64)
        self.add(self.zmm23)
        self.ymm23 = state.RegisterAlias("ymm23", self.zmm23, 32, 0)
        self.add(self.ymm23)
        self.xmm23 = state.RegisterAlias("xmm23", self.zmm23, 16, 0)
        self.add(self.xmm23)

        self.zmm24 = state.Register("zmm24", 64)
        self.add(self.zmm24)
        self.ymm24 = state.RegisterAlias("ymm24", self.zmm24, 32, 0)
        self.add(self.ymm24)
        self.xmm24 = state.RegisterAlias("xmm24", self.zmm24, 16, 0)
        self.add(self.xmm24)

        self.zmm25 = state.Register("zmm25", 64)
        self.add(self.zmm25)
        self.ymm25 = state.RegisterAlias("ymm25", self.zmm25, 32, 0)
        self.add(self.ymm25)
        self.xmm25 = state.RegisterAlias("xmm25", self.zmm25, 16, 0)
        self.add(self.xmm25)

        self.zmm26 = state.Register("zmm26", 64)
        self.add(self.zmm26)
        self.ymm26 = state.RegisterAlias("ymm26", self.zmm26, 32, 0)
        self.add(self.ymm26)
        self.xmm26 = state.RegisterAlias("xmm26", self.zmm26, 16, 0)
        self.add(self.xmm26)

        self.zmm27 = state.Register("zmm27", 64)
        self.add(self.zmm27)
        self.ymm27 = state.RegisterAlias("ymm27", self.zmm27, 32, 0)
        self.add(self.ymm27)
        self.xmm27 = state.RegisterAlias("xmm27", self.zmm27, 16, 0)
        self.add(self.xmm27)

        self.zmm28 = state.Register("zmm28", 64)
        self.add(self.zmm28)
        self.ymm28 = state.RegisterAlias("ymm28", self.zmm28, 32, 0)
        self.add(self.ymm28)
        self.xmm28 = state.RegisterAlias("xmm28", self.zmm28, 16, 0)
        self.add(self.xmm28)

        self.zmm29 = state.Register("zmm29", 64)
        self.add(self.zmm29)
        self.ymm29 = state.RegisterAlias("ymm29", self.zmm29, 32, 0)
        self.add(self.ymm29)
        self.xmm29 = state.RegisterAlias("xmm29", self.zmm29, 16, 0)
        self.add(self.xmm29)

        self.zmm30 = state.Register("zmm30", 64)
        self.add(self.zmm30)
        self.ymm30 = state.RegisterAlias("ymm30", self.zmm30, 32, 0)
        self.add(self.ymm30)
        self.xmm30 = state.RegisterAlias("xmm30", self.zmm30, 16, 0)
        self.add(self.xmm30)

        self.zmm31 = state.Register("zmm31", 64)
        self.add(self.zmm31)
        self.ymm31 = state.RegisterAlias("ymm31", self.zmm31, 32, 0)
        self.add(self.ymm31)
        self.xmm31 = state.RegisterAlias("xmm31", self.zmm31, 16, 0)
        self.add(self.xmm31)



from .aarch64 import AArch64MachineDef
from .amd64 import AMD64MachineDef
from .arm import (
    ARMv5TMachineDef,
    ARMv6MMachineDef,
    ARMv6MThumbMachineDef,
    ARMv7MMachineDef,
)
from .i386 import i386MachineDef
from .loongarch import LoongArch64MachineDef
from .machdef import AngrMachineDef
from .mips import MIPSBEMachineDef, MIPSELMachineDef
from .mips64 import MIPS64BEMachineDef, MIPS64ELMachineDef
from .ppc import PowerPC32MachineDef, PowerPC64MachineDef
from .riscv import RISCV64MachineDef
from .xtensa import XTensaBEMachineDef, XTensaELMachineDef

__all__ = [
    "AArch64MachineDef",
    "AMD64MachineDef",
    "AngrMachineDef",
    "ARMv5TMachineDef",
    "ARMv6MMachineDef",
    "ARMv6MThumbMachineDef",
    "ARMv7MMachineDef",
    "i386MachineDef",
    "LoongArch64MachineDef",
    "MIPSBEMachineDef",
    "MIPSELMachineDef",
    "MIPS64BEMachineDef",
    "MIPS64ELMachineDef",
    "PowerPC32MachineDef",
    "PowerPC64MachineDef",
    "RISCV64MachineDef",
    "XTensaBEMachineDef",
    "XTensaELMachineDef",
]



from ....platforms import Architecture, Byteorder
from .machdef import GhidraMachineDef


class MIPSMachineDef(GhidraMachineDef):
    arch = Architecture.MIPS32

    # NOTE: MIPS registers have a name and a number
    # angr's machine state doesn't use the number,
    # so... name.
    _registers = {
        # *** General-Purpose Registers ***
        # Assembler-Temporary Register
        "at": "at",
        "1": "at",
        # Return Value Registers
        "v0": "v0",
        "2": "v0",
        "v1": "v1",
        "3": "v1",
        # Argument Registers
        "a0": "a0",
        "4": "a0",
        "a1": "a1",
        "5": "a1",
        "a2": "a2",
        "6": "a2",
        "a3": "a3",
        "7": "a3",
        # Temporary Registers
        "t0": "t0",
        "8": "t0",
        "t1": "t1",
        "9": "t1",
        "t2": "t2",
        "10": "t2",
        "t3": "t3",
        "11": "t3",
        "t4": "t4",
        "12": "t4",
        "t5": "t5",
        "13": "t5",
        "t6": "t6",
        "14": "t6",
        "t7": "t7",
        "15": "t7",
        # NOTE: These numbers aren't out of order.
        # t8 and t9 are later in the register file than t0 - t7.
        "t8": "t8",
        "24": "t8",
        "t9": "t9",
        "25": "t9",
        # Saved Registers
        "s0": "s0",
        "16": "s0",
        "s1": "s1",
        "17": "s1",
        "s2": "s2",
        "18": "s2",
        "s3": "s3",
        "19": "s3",
        "s4": "s4",
        "20": "s4",
        "s5": "s5",
        "21": "s5",
        "s6": "s6",
        "22": "s6",
        "s7": "s7",
        "23": "s7",
        # NOTE: Register #30 was originally the Frame Pointer.
        # It's been re-aliased as s8, since many ABIs don't use the frame pointer.
        # Unicorn and Sleigh prefer to use the alias s8,
        # so it should be the base register.
        "s8": "s8",
        "fp": "s8",
        "30": "s8",
        # Kernel-reserved Registers
        "k0": "k0",
        "26": "k0",
        "k1": "k1",
        "27": "k1",
        # *** Pointer Registers ***
        # Zero register
        "zero": "zero",
        "0": "zero",
        # Global Offset Pointer
        "gp": "gp",
        "28": "gp",
        # Stack Pointer
        "sp": "sp",
        "29": "sp",
        # Return Address
        "ra": "ra",
        "31": "ra",
        # Program Counter
        "pc": "pc",
        # Floating Point Registers
        "f0": "f0",
        "f1": "f1",
        "f2": "f2",
        "f3": "f3",
        "f4": "f4",
        "f5": "f5",
        "f6": "f6",
        "f7": "f7",
        "f8": "f8",
        "f9": "f9",
        "f10": "f10",
        "f11": "f11",
        "f12": "f12",
        "f13": "f13",
        "f14": "f14",
        "f15": "f15",
        "f16": "f16",
        "f17": "f17",
        "f18": "f18",
        "f19": "f19",
        "f20": "f20",
        "f21": "f21",
        "f22": "f22",
        "f23": "f23",
        "f24": "f24",
        "f25": "f25",
        "f26": "f26",
        "f27": "f27",
        "f28": "f28",
        "f29": "f29",
        "f30": "f30",
        "f31": "f31",
        # *** Floating Point Control Registers ***
        "fir": "fir",
        "fcsr": "fcsr",
        "fexr": "fexr",
        "fenr": "fenr",
        "fccr": "fccr",
        # *** Accumulator Registers ***
        # MIPS uses these to implement 64-bit results
        # from 32-bit multiplication, amongst others.
        "ac0": None,
        "hi0": None,
        "lo0": None,
        "ac1": None,
        "hi1": None,
        "lo1": None,
        "ac2": None,
        "hi2": None,
        "lo2": None,
        "ac3": None,
        "hi3": None,
        "lo3": None,
    }

    supports_single_step = True


class MIPSELMachineDef(MIPSMachineDef):
    byteorder = Byteorder.LITTLE
    language_id = "MIPS:LE:32:default"


class MIPSBEMachineDef(MIPSMachineDef):
    byteorder = Byteorder.BIG
    language_id = "MIPS:BE:32:default"



import typing

import capstone

from ..platforms import Architecture, Byteorder
from .platformdef import PlatformDef, RegisterAliasDef, RegisterDef

# NOTE: mips32 has two major ISA variants.
#
# The Debian standard is built around mips32r2.
#
# There's a newer specification mips32r6 that adds and deprecates
# a number of instructions, amongst other changes.
#
# There are older ISA variants that are still in use
# because no one updates their hardware.
#
# This definition will focus on mips32r2
# If you need support for r6, please submit a ticket.

# NOTE: MIPS registers don't really have canonical uses, or canonical names.
# Their names are assigned based on the purpose they serve
# in a specific ABI.
#
# Thus far, I've found two mips32 ABIs:
#
# o32 is the original ABI.  It assigns four argument registers.
#
# n32 is a back-port of the n64 ABI, which assigns eight argument registers.
# This one's not super-popular.


class MIPSO32PlatformDef(PlatformDef):
    # Abstract MIPS platform definition based on the o32 ABI.
    architecture = Architecture.MIPS32

    address_size = 4

    capstone_arch = capstone.CS_ARCH_MIPS
    capstone_mode = capstone.CS_MODE_MIPS32

    conditional_branch_mnemonics = {
        # Conditional branch
        "beq",
        "beqz",
        "bne",
        "bnez",
        "bgez",
        "bgtz",
        "blez",
        "bltz",
        # Conditional branch-and-link
        "bgezal",
        "bltzal"
        # Likely conditional branch
        # Skip the delay slot if they are not taken.
        "beql",
        "bnel",
        "bgezl",
        "bgtzl",
        "blezl",
        "bltzl",
        # Likely conditional branch-and-link
        # Skip the delay slot if they are not taken
        "bgezall",
        "bltzall",
    }

    compare_mnemonics = {
        # MIPS doesn't really have integer comparison instructions
        # All of the conditional branches include a comparsion
        # relative to zero; the compiler needs to reduce
        # all conditional tests to comparisons against zero.
        # Floating-point comparison
        # Save to FCC
        # NOTE: Unlike branches, compares only support eq, lt, and le
        "c.eq.s",
        "c.eq.d",
        "c.eq.ps",
        "c.lt.s",
        "c.lt.d",
        "c.lt.ps",
        "c.le.s",
        "c.le.d",
        "c.le.ps",
        # Floating-point comparison
        # Save to FPR
        "cmp.eq.s",
        "cmp.eq.d",
        "cmp.eq.ps",
        "cmp.lt.s",
        "cmp.lt.d",
        "cmp.lt.ps",
        "cmp.le.s",
        "cmp.le.d",
        "cmp.le.ps",
    }

    delay_slot_mnemonics = {
        "b",
        "bal",
        "j",
        "jr",
        "jal",
        "jalr",
    } | conditional_branch_mnemonics

    pc_register = "pc"
    sp_register = "sp"

    # Special registers
    # zero: Hard-wired to zero
    # at: Reserved for assembler
    # sp: Stack pointer
    # gp: Global pointer
    # kX: Reserved for kernel on most platforms
    # fX: Floating-point registers
    # aX: Accumulator registers used in multiplication
    general_purpose_registers = [
        "v0",
        "v1",
        "a0",
        "a1",
        "a2",
        "a3",
        "t0",
        "t1",
        "t2",
        "t3",
        "t4",
        "t5",
        "t6",
        "t7",
        "t8",
        "t9",
        "s0",
        "s1",
        "s2",
        "s3",
        "s4",
        "s5",
        "s6",
        "s7",
        "s8",
    ]

    @property
    def registers(self) -> typing.Dict[str, RegisterDef]:
        return self._registers

    def __init__(self):
        # For analysis purposes, the
        self._registers = {
            # *** General-Purpose Registers ***
            # Assembler-Temporary Register
            "at": RegisterDef(name="at", size=4),
            "1": RegisterAliasDef(name="1", parent="at", size=4, offset=0),
            # Return Value Registers
            "v0": RegisterDef(name="v0", size=4),
            "2": RegisterAliasDef(name="2", parent="v0", size=4, offset=0),
            "v1": RegisterDef(name="v1", size=4),
            "3": RegisterAliasDef(name="3", parent="v1", size=4, offset=0),
            # Argument Registers
            "a0": RegisterDef(name="a0", size=4),
            "4": RegisterAliasDef(name="4", parent="a0", size=4, offset=0),
            "a1": RegisterDef(name="a1", size=4),
            "5": RegisterAliasDef(name="5", parent="a1", size=4, offset=0),
            "a2": RegisterDef(name="a2", size=4),
            "6": RegisterAliasDef(name="6", parent="a2", size=4, offset=0),
            "a3": RegisterDef(name="a3", size=4),
            "7": RegisterAliasDef(name="7", parent="a3", size=4, offset=0),
            # Temporary Registers
            "t0": RegisterDef(name="t0", size=4),
            "8": RegisterAliasDef(name="8", parent="t0", size=4, offset=0),
            "t1": RegisterDef(name="t1", size=4),
            "9": RegisterAliasDef(name="9", parent="t1", size=4, offset=0),
            "t2": RegisterDef(name="t2", size=4),
            "10": RegisterAliasDef(name="10", parent="t2", size=4, offset=0),
            "t3": RegisterDef(name="t3", size=4),
            "11": RegisterAliasDef(name="11", parent="t3", size=4, offset=0),
            "t4": RegisterDef(name="t4", size=4),
            "12": RegisterAliasDef(name="12", parent="t4", size=4, offset=0),
            "t5": RegisterDef(name="t5", size=4),
            "13": RegisterAliasDef(name="13", parent="t5", size=4, offset=0),
            "t6": RegisterDef(name="t6", size=4),
            "14": RegisterAliasDef(name="14", parent="t6", size=4, offset=0),
            "t7": RegisterDef(name="t7", size=4),
            "15": RegisterAliasDef(name="15", parent="t7", size=4, offset=0),
            # NOTE: These numbers aren't out of order.
            # t8 and t9 are later in the register file than t0 - t7.
            "t8": RegisterDef(name="t8", size=4),
            "24": RegisterAliasDef(name="24", parent="t8", size=4, offset=0),
            "t9": RegisterDef(name="t9", size=4),
            "25": RegisterAliasDef(name="25", parent="t9", size=4, offset=0),
            # Saved Registers
            "s0": RegisterDef(name="s0", size=4),
            "16": RegisterAliasDef(name="16", parent="s0", size=4, offset=0),
            "s1": RegisterDef(name="s1", size=4),
            "17": RegisterAliasDef(name="17", parent="s1", size=4, offset=0),
            "s2": RegisterDef(name="s2", size=4),
            "18": RegisterAliasDef(name="18", parent="s2", size=4, offset=0),
            "s3": RegisterDef(name="s3", size=4),
            "19": RegisterAliasDef(name="19", parent="s3", size=4, offset=0),
            "s4": RegisterDef(name="s4", size=4),
            "20": RegisterAliasDef(name="20", parent="s4", size=4, offset=0),
            "s5": RegisterDef(name="s5", size=4),
            "21": RegisterAliasDef(name="21", parent="s5", size=4, offset=0),
            "s6": RegisterDef(name="s6", size=4),
            "22": RegisterAliasDef(name="22", parent="s6", size=4, offset=0),
            "s7": RegisterDef(name="s7", size=4),
            "23": RegisterAliasDef(name="23", parent="s7", size=4, offset=0),
            # NOTE: Register #30 was originally the Frame Pointer.
            # It's been re-aliased as s8, since many ABIs don't use the frame pointer.
            # Unicorn and Sleigh prefer to use the alias s8,
            # so it should be the base register.
            "s8": RegisterDef(name="s8", size=4),
            "fp": RegisterAliasDef(name="fp", parent="s8", size=4, offset=0),
            "30": RegisterAliasDef(name="30", parent="s8", size=4, offset=0),
            # Kernel-reserved Registers
            "k0": RegisterDef(name="k0", size=4),
            "26": RegisterAliasDef(name="26", parent="k0", size=4, offset=0),
            "k1": RegisterDef(name="k1", size=4),
            "27": RegisterAliasDef(name="27", parent="k1", size=4, offset=0),
            # *** Pointer Registers ***
            # Zero register
            "zero": RegisterDef(name="zero", size=4),
            "0": RegisterAliasDef(name="0", parent="zero", size=4, offset=0),
            # Global Offset Pointer
            "gp": RegisterDef(name="gp", size=4),
            "28": RegisterAliasDef(name="28", parent="gp", size=4, offset=0),
            # Stack Pointer
            "sp": RegisterDef(name="sp", size=4),
            "29": RegisterAliasDef(name="29", parent="sp", size=4, offset=0),
            # Return Address
            "ra": RegisterDef(name="ra", size=4),
            "31": RegisterAliasDef(name="31", parent="ra", size=4, offset=0),
            # Program Counter
            "pc": RegisterDef(name="pc", size=4),
            # NOTE: MIPS has no ALU flags or status register
            # It doesn't have a non-privileged status register,
            # and the conditional tests write the result to a general register.
            # *** Floating Point Registers ***
            "f0": RegisterDef(name="f0", size=8),
            "f1": RegisterDef(name="f1", size=8),
            "f2": RegisterDef(name="f2", size=8),
            "f3": RegisterDef(name="f3", size=8),
            "f4": RegisterDef(name="f4", size=8),
            "f5": RegisterDef(name="f5", size=8),
            "f6": RegisterDef(name="f6", size=8),
            "f7": RegisterDef(name="f7", size=8),
            "f8": RegisterDef(name="f8", size=8),
            "f9": RegisterDef(name="f9", size=8),
            "f10": RegisterDef(name="f10", size=8),
            "f11": RegisterDef(name="f11", size=8),
            "f12": RegisterDef(name="f12", size=8),
            "f13": RegisterDef(name="f13", size=8),
            "f14": RegisterDef(name="f14", size=8),
            "f15": RegisterDef(name="f15", size=8),
            "f16": RegisterDef(name="f16", size=8),
            "f17": RegisterDef(name="f17", size=8),
            "f18": RegisterDef(name="f18", size=8),
            "f19": RegisterDef(name="f19", size=8),
            "f20": RegisterDef(name="f20", size=8),
            "f21": RegisterDef(name="f21", size=8),
            "f22": RegisterDef(name="f22", size=8),
            "f23": RegisterDef(name="f23", size=8),
            "f24": RegisterDef(name="f24", size=8),
            "f25": RegisterDef(name="f25", size=8),
            "f26": RegisterDef(name="f26", size=8),
            "f27": RegisterDef(name="f27", size=8),
            "f28": RegisterDef(name="f28", size=8),
            "f29": RegisterDef(name="f29", size=8),
            "f30": RegisterDef(name="f30", size=8),
            "f31": RegisterDef(name="f31", size=8),
            # *** Floating Point Control Registers ***
            "fir": RegisterDef(name="fir", size=4),
            "fcsr": RegisterDef(name="fcsr", size=4),
            "fexr": RegisterDef(name="fexr", size=4),
            "fenr": RegisterDef(name="fenr", size=4),
            "fccr": RegisterDef(name="fccr", size=4),
            # TODO: MIPS has a boatload of extensions with their own registers.
        }


class MIPS32EL(MIPSO32PlatformDef):
    byteorder = Byteorder.LITTLE

    def __init__(self):
        super().__init__()
        self._registers |= {
            # *** Accumulator Registers ***
            # MIPS uses these to implement 64-bit results
            # from 32-bit multiplication, amongst others.
            "ac0": RegisterDef(name="ac0", size=8),
            "lo0": RegisterAliasDef(name="lo0", parent="ac0", size=4, offset=0),
            "hi0": RegisterAliasDef(name="hi0", parent="ac0", size=4, offset=4),
            "ac1": RegisterDef(name="ac1", size=8),
            "lo1": RegisterAliasDef(name="lo1", parent="ac1", size=4, offset=0),
            "hi1": RegisterAliasDef(name="hi1", parent="ac1", size=4, offset=4),
            "ac2": RegisterDef(name="ac2", size=8),
            "lo2": RegisterAliasDef(name="lo2", parent="ac2", size=4, offset=0),
            "hi2": RegisterAliasDef(name="hi2", parent="ac2", size=4, offset=4),
            "ac3": RegisterDef(name="ac3", size=8),
            "lo3": RegisterAliasDef(name="lo3", parent="ac3", size=4, offset=0),
            "hi3": RegisterAliasDef(name="hi3", parent="ac3", size=4, offset=4),
        }


class MIPS32BE(MIPSO32PlatformDef):
    byteorder = Byteorder.BIG

    capstone_mode = capstone.CS_MODE_MIPS32 | capstone.CS_MODE_BIG_ENDIAN

    def __init__(self):
        super().__init__()
        self._registers |= {
            # *** Accumulator Registers ***
            # MIPS uses these to implement 64-bit results
            # from 32-bit multiplication, amongst others.
            "ac0": RegisterDef(name="ac0", size=8),
            "hi0": RegisterAliasDef(name="hi0", parent="ac0", size=4, offset=0),
            "lo0": RegisterAliasDef(name="lo0", parent="ac0", size=4, offset=4),
            "ac1": RegisterDef(name="ac1", size=8),
            "hi1": RegisterAliasDef(name="hi1", parent="ac1", size=4, offset=0),
            "lo1": RegisterAliasDef(name="lo1", parent="ac1", size=4, offset=4),
            "ac2": RegisterDef(name="ac2", size=8),
            "hi2": RegisterAliasDef(name="hi2", parent="ac2", size=4, offset=0),
            "lo2": RegisterAliasDef(name="lo2", parent="ac2", size=4, offset=4),
            "ac3": RegisterDef(name="ac3", size=8),
            "hi3": RegisterAliasDef(name="hi3", parent="ac3", size=4, offset=0),
            "lo3": RegisterAliasDef(name="lo3", parent="ac3", size=4, offset=4),
        }



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import MIPSSysVModel


class MIPSSysVTime(Time, MIPSSysVModel):
    pass


class MIPSSysVLocaltime(Localtime, MIPSSysVModel):
    pass


class MIPSSysVGmtime(Gmtime, MIPSSysVModel):
    pass


class MIPSSysVCtime(Ctime, MIPSSysVModel):
    pass


class MIPSSysVAsctime(Asctime, MIPSSysVModel):
    pass


class MIPSSysVStrftime(Strftime, MIPSSysVModel):
    pass


class MIPSSysVDifftime(Difftime, MIPSSysVModel):
    pass


class MIPSSysVMktime(Mktime, MIPSSysVModel):
    pass


class MIPSSysVClock(Clock, MIPSSysVModel):
    pass


__all__ = [
    "MIPSSysVTime",
    "MIPSSysVLocaltime",
    "MIPSSysVGmtime",
    "MIPSSysVCtime",
    "MIPSSysVAsctime",
    "MIPSSysVStrftime",
    "MIPSSysVDifftime",
    "MIPSSysVMktime",
    "MIPSSysVClock",
]



import logging
import typing

import claripy

from ... import emulators, exceptions, platforms, state
from ...emulators.angr.exceptions import PathTerminationSignal

log = logging.getLogger(__name__)


class MallocModel(state.models.Model):
    name = "malloc"
    platform = platforms.Platform(
        platforms.Architecture.X86_64, platforms.Byteorder.LITTLE
    )
    abi = platforms.ABI.NONE

    _arg1_for_arch = {
        platforms.Architecture.AARCH64: "x0",
        platforms.Architecture.ARM_V5T: "r0",
        platforms.Architecture.ARM_V7A: "r0",
        platforms.Architecture.MIPS32: "a0",
        platforms.Architecture.MIPS64: "a0",
        platforms.Architecture.POWERPC32: "r3",
        platforms.Architecture.X86_64: "rdi",
    }

    _ret_for_arch = {
        platforms.Architecture.AARCH64: "x0",
        platforms.Architecture.ARM_V5T: "r0",
        platforms.Architecture.ARM_V7A: "r0",
        platforms.Architecture.MIPS32: "v0",
        platforms.Architecture.MIPS64: "v0",
        platforms.Architecture.POWERPC32: "r3",
        platforms.Architecture.X86_64: "rax",
    }

    def __init__(
        self,
        address: int,
        heap: state.memory.heap.Heap,
        platform: platforms.Platform,
        read_callback,
        write_callback,
    ):
        super().__init__(address)

        self.arg1_reg = self._arg1_for_arch[platform.architecture]
        self.ret_reg = self._ret_for_arch[platform.architecture]

        if len(heap) > 0:
            raise exceptions.ConfigurationError("This only works with a blank heap")
        end = state.IntegerValue(heap.address + heap.get_capacity(), 8, None)
        ptr = state.IntegerValue(heap.address + 16, 8, None)

        heap.allocate(end)
        heap.allocate(ptr)

        self.heap_addr = heap.address
        self.read_callback = read_callback
        self.write_callback = write_callback

        self.struct_lengths: typing.Dict[str, int] = dict()
        self.struct_prefixes: typing.Dict[str, str] = dict()
        self.struct_fields: typing.Dict[
            str, typing.List[typing.Tuple[int, str]]
        ] = dict()

    def bind_length_to_struct(
        self, field: str, prefix: str, labels: typing.List[typing.Tuple[int, str]]
    ):
        self.struct_lengths[field] = 0
        self.struct_prefixes[field] = prefix
        for size, label in labels:
            self.struct_lengths[field] += size
        self.struct_fields[field] = labels

    def model(self, emulator: emulators.Emulator):
        if not isinstance(emulator, emulators.AngrEmulator):
            raise exceptions.ConfigurationError("Model only works with angr")
        # Read the capacity.
        # Bypass the smallworld API; I want to know if it's symbolic
        fda = emulator.get_extension("fda")
        capacity = emulator.read_register_symbolic(self.arg1_reg)
        length = 0

        if capacity.symbolic:
            # Capacity is a symbolic expression.
            # Check if it's collapsible
            good = True
            try:
                vals = emulator.eval_atmost(capacity, 1)
            except exceptions.UnsatError:
                log.error("Capacity is unsat")
                good = False
            except exceptions.SymbolicValueError:
                log.error("Capacity did not collapse to one value")
                good = False

            if not good or len(vals) > 1:
                log.error("The following fields are lengths:")
                for var in capacity.variables:
                    log.error(f"   {var}")
                log.error("Concretize them to continue analysis")
                raise PathTerminationSignal()

            length = vals[0]
        else:
            length = capacity.concrete_value

        heap_end = int.from_bytes(
            emulator.read_memory_content(self.heap_addr, 8), "little"
        )
        heap_ptr = int.from_bytes(
            emulator.read_memory_content(self.heap_addr + 8, 8), "little"
        )

        if heap_end - heap_ptr < length:
            # OOM; Return NULL
            log.warning(
                f"malloc: OOM! Need {length}, only have {hex(heap_end)} - {hex(heap_ptr)} = {heap_end - heap_ptr}"
            )
            res = 0
        else:
            # Return the pointer to the new region.
            log.warning(f"malloc: Alloc'd {length} bytes at {hex(heap_ptr)}")
            res = heap_ptr
            heap_ptr += length
            emulator.write_memory_content(
                self.heap_addr + 8, heap_ptr.to_bytes(8, "little")
            )

            # Check if this is something we want to track.
            fields = list(map(lambda x: x.split("_")[0], capacity.variables))

            if len(fields) != 1:
                # Don't track; too complex
                log.warn("  Length has multiple variables; not solving")
            elif fields[0] not in self.struct_lengths:
                # Don't track; not something we asked to track
                log.warn(f"  Length field {fields[0]} not known")
            else:
                struct_length = self.struct_lengths[fields[0]]
                struct_prefix = self.struct_prefixes[fields[0]]
                struct_labels = self.struct_fields[fields[0]]

                sym = list(filter(lambda x: x.op == "BVS", capacity.leaf_asts()))[0]
                if not isinstance(sym, claripy.ast.bv.BV):
                    raise TypeError("BVS isn't a bitvector: {sym}")
                n = emulator.eval_atmost(sym, 1)[0]

                if length // n != struct_length:
                    # Don't track; not an obvious "n * sizeof(struct foo)"
                    log.warn(f"  {length} // {n} != {struct_length}")
                elif length % struct_length != 0:
                    # Don't track; not an obvious "n * sizeof(struct foo)"
                    log.warn(f"  {length} not evenly divisible by {n}")
                else:
                    # Track!
                    log.warn(f"  Alloc of {n} items")
                    addr = res
                    for i in range(0, n):
                        struct_fields = list()
                        off = 0
                        for flen, fname in struct_labels:
                            # Build a symbol for this field of this item
                            flabel = struct_prefix + "." + str(i) + "." + fname
                            expr = claripy.BVS(flabel, flen * 8, explicit_name=True)
                            struct_fields.append(expr)

                            # Track the new field
                            r = range(addr + off, addr + off + flen)
                            fda.fda_labels.add(flabel)
                            fda.fda_addr_to_label[r] = flabel
                            fda.fda_label_to_addr[flabel] = r
                            fda.fda_bindings[flabel] = expr

                            off += flen

                        # Insert the whole struct into memory
                        expr = claripy.Concat(*struct_fields)
                        emulator.write_memory(addr, expr)

                        addr += struct_length

                    # Hook the entire array
                    fda.fda_mem_ranges.add((res, res + length))
                    emulator.hook_memory_read_symbolic(
                        res, res + length, self.read_callback
                    )
                    emulator.hook_memory_write_symbolic(
                        res, res + length, self.write_callback
                    )

        emulator.write_register_content(self.ret_reg, res)


class FreeModel(state.models.Model):
    name = "free"
    platform = platforms.Platform(
        platforms.Architecture.X86_64, platforms.Byteorder.LITTLE
    )
    abi = platforms.ABI.NONE

    def __init__(self, address: int):
        super().__init__(address)

    def model(self, emulator: emulators.Emulator):
        # Just free it.
        return



from ..... import platforms
from .....exceptions import ConfigurationError
from ..structs import ElfRela
from .rela import ElfRelocator

R_LARCH_64 = 2  # Direct 64-bit
R_LARCH_RELATIVE = 4  # Adjust by program base
R_LARCH_JUMP_SLOT = 5  # Create PLT entry
R_LARCH_NUM = 127  # This and higher aren't valid


class LoongArch64ElfRelocator(ElfRelocator):
    arch = platforms.Architecture.LOONGARCH64
    byteorder = platforms.Byteorder.LITTLE

    def _compute_value(self, rela: ElfRela):
        if (
            rela.type == R_LARCH_JUMP_SLOT
            or rela.type == R_LARCH_RELATIVE
            or rela.type == R_LARCH_64
        ):
            # Different semantics, all behave the same
            val = rela.symbol.value + rela.symbol.baseaddr + rela.addend
            return val.to_bytes(8, "little")
        elif rela.type >= 0 and rela.type < R_LARCH_NUM:
            raise ConfigurationError(
                f"Valid, but unsupported relocation for {rela.symbol.name}: {rela.type}"
            )
        else:
            raise ConfigurationError(
                f"Invalid relocation type for {rela.symbol.name}: {rela.type}"
            )



from angr.storage.memory_mixins.memory_mixin import MemoryMixin


class FixupMemoryMixin(MemoryMixin):
    """Mixin providing some basic bug-fixed/feature improvements

    angr doesn't always behave as I'd expect.
    This mixin helps things a little.
    """

    def _concretize_addr(self, supercall, addr, strategies=None, condition=None):
        # Handle potential errors durring concretization
        # This works the same for reads and writes, so why duplicate code?

        if not self.state.solver.satisfiable():
            # This state isn't satisfiable; concretization is impossible, but irrelevant.
            #
            # Under some conditions of which I'm uncertain,
            # the concretizer will try to run even if the state is already unsat.
            # Since the concretizer relies on the solver,
            # it fails with a big ole' fatal exception.
            #
            # The successor computation will eventually kill these states anyway,
            # so just return a dummy value to keep things running.
            return [0xDEAD0000]

        return supercall(addr, strategies=strategies, condition=condition)

    def concretize_read_addr(self, addr, strategies=None, condition=None):
        return self._concretize_addr(
            super().concretize_read_addr,
            addr,
            strategies=strategies,
            condition=condition,
        )

    def concretize_write_addr(self, addr, strategies=None, condition=None):
        return self._concretize_addr(
            super().concretize_write_addr,
            addr,
            strategies=strategies,
            condition=condition,
        )



import typing

import lief

from ...... import platforms
from .prstatus import PrStatus


class PowerPC(PrStatus):
    byteorder = platforms.Byteorder.BIG

    @property
    def register_coords(
        self,
    ) -> typing.List[typing.Tuple[typing.Optional[str], int, int]]:
        return self._register_coords

    def __init__(self, elf: lief.ELF.Binary, wordsize):
        _register_list = [
            "r0",
            "r1",
            "r2",
            "r3",
            "r4",
            "r5",
            "r6",
            "r7",
            "r8",
            "r9",
            "r10",
            "r11",
            "r12",
            "r13",
            "r14",
            "r15",
            "r16",
            "r17",
            "r18",
            "r19",
            "r20",
            "r21",
            "r22",
            "r23",
            "r24",
            "r25",
            "r26",
            "r27",
            "r28",
            "r29",
            "r30",
            "r31",
            "pc",
            # No idea what these are, as it's not documented.
            None,
            None,
            "ctr",
            "lr",
        ]
        self._register_coords = [
            (_register_list[i], i * wordsize, wordsize)
            for i in range(0, len(_register_list))
        ]
        super().__init__(elf)


class PowerPC32(PowerPC):
    architecture = platforms.Architecture.POWERPC32
    pr_regs_off = 72
    pr_regs_size = 192

    def __init__(self, elf: lief.ELF.Binary):
        super().__init__(elf, 4)


class PowerPC64(PowerPC):
    architecture = platforms.Architecture.POWERPC64

    pr_regs_off = 112
    pr_regs_size = 384

    def __init__(self, elf: lief.ELF.Binary):
        super().__init__(elf, 8)



import json
import typing
from dataclasses import dataclass
from enum import Enum

from smallworld.instructions import Operand

CmpInfo = typing.Tuple[str, Operand, int]


class TraceRes(Enum):
    ER_NONE = 0
    ER_BOUNDS = 1
    ER_MAX_INSNS = 2
    ER_FAIL = 3


# one element in a trace
@dataclass
class TraceElement:
    pc: int
    ic: int  # instruction count
    mnemonic: str
    op_str: str
    cmp: typing.List[CmpInfo]
    branch: bool
    immediates: typing.List[int]

    def __str__(self):
        return f"{self.ic} 0x{self.pc:x} [{self.mnemonic} {self.op_str}] {self.cmp} {self.branch} {self.immediates}"

    # NOTE this is just used for logging. Can't un-jsonify
    def to_json(self):
        d = {
            "instruction_count": self.ic,
            "pc": self.pc,
            "instr": f"{self.mnemonic} {self.op_str}",
            "cmp": str(self.cmp),
            "branch": self.branch,
            "immediates": self.immediates,
        }
        return json.dumps(d)



import typing

import angr
import archinfo
import pyvex

from ....platforms import Architecture, Byteorder
from .machdef import AngrMachineDef


class i386MachineDef(AngrMachineDef):
    arch = Architecture.X86_32
    byteorder = Byteorder.LITTLE

    angr_arch = archinfo.arch_x86.ArchX86()

    _registers = {
        # *** General Purpose Registers ***
        "eax": "eax",
        "ax": "ax",
        "al": "al",
        "ah": "ah",
        "ebx": "ebx",
        "bx": "bx",
        "bl": "bl",
        "bh": "bh",
        "ecx": "ecx",
        "cx": "cx",
        "cl": "cl",
        "ch": "ch",
        "edx": "edx",
        "dx": "dx",
        "dl": "dl",
        "dh": "dh",
        "esi": "esi",
        "si": "si",
        "sil": "sil",
        "edi": "edi",
        "di": "di",
        "dil": "dil",
        "ebp": "ebp",
        "bp": "bp",
        "bpl": "bpl",
        "esp": "esp",
        "sp": "sp",
        "spl": "spl",
        # *** Instruction Pointer ***
        "eip": "eip",
        "ip": "ip",
        # *** Segment Registers ***
        "cs": "cs",
        "ds": "ds",
        "es": "es",
        "fs": "fs",
        "gs": "gs",
        "ss": "ss",
        # *** Flags Register ***
        "eflags": "eflags",
        "flags": "flags",
        # *** Control Registers ***
        "cr0": "",
        "cr1": "",
        "cr2": "",
        "cr3": "",
        "cr4": "",
        "cr8": "",
        # *** Debug Registers ***
        "dr0": "",
        "dr1": "",
        "dr2": "",
        "dr3": "",
        "dr6": "",
        "dr7": "",
        # *** Descriptor Table Registers ***
        "gdtr": "gdt",
        "idtr": "idt",
        "ldtr": "ldt",
        # *** Task Register ***
        "tr": "",
        # *** x87 Registers ***
        # TODO: angr seems to support x87, but I have no idea how its register file works
        # I can't find most of the control registers,
        # and there don't seem to be separate "fprN" registers; just one giant blob
        "fpr0": "",
        "fpr1": "",
        "fpr2": "",
        "fpr3": "",
        "fpr4": "",
        "fpr5": "",
        "fpr6": "",
        "fpr7": "",
        "fctrl": "",
        "fstat": "",
        "ftag": "fptag",
        "fip": "",
        "fdp": "",
        "fop": "",
        # *** MMX Registers ***
        "mm0": "mm0",
        "mm1": "mm1",
        "mm2": "mm2",
        "mm3": "mm3",
        "mm4": "mm4",
        "mm5": "mm5",
        "mm6": "mm6",
        "mm7": "mm7",
        # *** SSE Registers ***
        "xmm0": "xmm0",
        "xmm1": "xmm1",
        "xmm2": "xmm2",
        "xmm3": "xmm3",
        "xmm4": "xmm4",
        "xmm5": "xmm5",
        "xmm6": "xmm6",
        "xmm7": "xmm7",
    }

    def rebuild_irsb(self, addr, good_bytes):
        # Lift good_bytes to an IRSB
        return pyvex.lift(good_bytes, addr, self.angr_arch)

    def build_sysexit_irsb(self, insn, good_bytes):
        # Build a new IRSB just containing sysexit:
        # 0: IMARK (addr, 2, 0)
        # 1: t0 = GET:I32(ecx)
        # 2: t1 = GET:I32(edx)
        # 3: PUT(esp) = t0
        # NEXT: PUT(eip) = t1; Iji_Sys_sysexit

        # Get the offsetfs of ecd and edx
        ecx_off, _ = self.angr_arch.registers["ecx"]
        edx_off, _ = self.angr_arch.registers["edx"]
        esp_off, _ = self.angr_arch.registers["esp"]

        # Type environment; we need two variables of type int32
        irsb_tyenv = pyvex.IRTypeEnv(self.angr_arch, ["Ity_I32", "Ity_I32"])

        # Statements
        irsb_stmts = [
            # Statement 0: Instruction mark
            pyvex.stmt.IMark(insn.address, 2, 0),
            # Statement 1: Load ecx into t0
            pyvex.stmt.WrTmp(
                0,  # Load into t0
                pyvex.expr.Get(  # Get a register
                    ecx_off,
                    "Ity_I32",
                ),
            ),
            # Statement 2: Load edx into t1
            pyvex.stmt.WrTmp(
                1,  # Load into t1
                pyvex.expr.Get(  # Get a register
                    edx_off,
                    "Ity_I32",
                ),
            ),
            # Statement 3: Load t0 into esp
            pyvex.stmt.Put(pyvex.expr.RdTmp.get_instance(0), esp_off),
        ]

        # "next" expression; Value of t1
        irsb_next = pyvex.expr.RdTmp.get_instance(1)
        # Jump kind
        irsb_jumpkind = "Ijk_Boring"

        irsb = pyvex.IRSB.from_py(
            irsb_tyenv,
            irsb_stmts,
            irsb_next,
            irsb_jumpkind,
            insn.address,
            self.angr_arch,
        )
        if len(good_bytes) > 0:
            prefix = self.rebuild_irsb(insn.address - len(good_bytes), good_bytes)
            # Fuse the two together
            irsb = prefix.extend(irsb)

        return irsb

    def successors(self, state: angr.SimState, **kwargs) -> typing.Any:
        # VEX doesn't correctly model SYSENTER and SYSEXIT

        # Fetch or compute the IR block for our state
        if "irsb" in kwargs and kwargs["irsb"] is not None:
            # Someone's already specified an IR block.
            irsb = kwargs["irsb"]
        else:
            # Disable optimization; it doesn't work

            # Compute the block from the state
            # Pray to the Powers that kwargs are compatible.
            irsb = state.block(**kwargs).vex

        if irsb.jumpkind == "Ijk_NoDecode":
            # VEX is stumped regarding this instruction.
            # Unfortunately, this also means basic block detection failed.
            irsb = None
            disas = state.block(**kwargs).disassembly
            good_bytes = b""
            for insn in disas.insns:
                if insn.mnemonic == "sysexit":
                    irsb = self.build_sysexit_irsb(insn, good_bytes)
                    break
                else:
                    good_bytes += insn.insn.bytes
            if irsb is None:
                irsb = self.rebuild_irsb(disas.insns[0].address, good_bytes)

        # Force the engine to use our IR block
        kwargs["irsb"] = irsb

        # Turn the crank on the engine
        try:
            return super().successors(state, **kwargs)
        except angr.errors.SimIRSBNoDecodeError as e:
            print(f"Bad IRSB:\n{irsb}")
            raise e



from ...exceptions import EmulationStop
from .underlay import AnalysisUnderlay


class BasicAnalysisUnderlay(AnalysisUnderlay):
    """Basic analysis underlay that steps until done"""

    def execute(self):
        try:
            while True:
                self.emulator.step()
        except EmulationStop:
            pass



import abc

from smallworld.hinting.hinting import Hinter

from .. import state, utils


class Analysis(utils.MetadataMixin):
    """An analysis that emits some information about some code, possibly to help with harnessing."""

    def __init__(self, hinter: Hinter, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.hinter = hinter

    @abc.abstractmethod
    def run(self, machine: state.Machine) -> None:
        """Run the analysis.

        This function **should not** modify the provided Machine. Instead, it
        should be coppied before modification.

        Arguments:
            machine: A machine state object on which this analysis should run.
        """

        pass


__all__ = ["Analysis"]



import typing

from .... import platforms
from .machdef import GhidraMachineDef


class AMD64MachineDef(GhidraMachineDef):
    arch: platforms.Architecture = platforms.Architecture.X86_64
    byteorder: platforms.Byteorder = platforms.Byteorder.LITTLE
    language_id: str = "x86:LE:64:default"

    _registers: typing.Dict[str, typing.Optional[str]] = {
        # *** General Purpose Registers ***
        "rax": "rax",
        "eax": "eax",
        "ax": "ax",
        "al": "al",
        "ah": "ah",
        "rbx": "rbx",
        "ebx": "ebx",
        "bx": "bx",
        "bl": "bl",
        "bh": "bh",
        "rcx": "rcx",
        "ecx": "ecx",
        "cx": "cx",
        "cl": "cl",
        "ch": "ch",
        "rdx": "rdx",
        "edx": "edx",
        "dx": "dx",
        "dl": "dl",
        "dh": "dh",
        "r8": "r8",
        "r8d": "r8d",
        "r8w": "r8w",
        "r8b": "r8b",
        "r9": "r9",
        "r9d": "r9d",
        "r9w": "r9w",
        "r9b": "r9b",
        "r10": "r10",
        "r10d": "r10d",
        "r10w": "r10w",
        "r10b": "r10b",
        "r11": "r11",
        "r11d": "r11d",
        "r11w": "r11w",
        "r11b": "r11b",
        "r12": "r12",
        "r12d": "r12d",
        "r12w": "r12w",
        "r12b": "r12b",
        "r13": "r13",
        "r13d": "r13d",
        "r13w": "r13w",
        "r13b": "r13b",
        "r14": "r14",
        "r14d": "r14d",
        "r14w": "r14w",
        "r14b": "r14b",
        "r15": "r15",
        "r15d": "r15d",
        "r15w": "r15w",
        "r15b": "r15b",
        "rsi": "rsi",
        "esi": "esi",
        "si": "si",
        "sil": "sil",
        "rdi": "rdi",
        "edi": "edi",
        "di": "di",
        "dil": "dil",
        "rbp": "rbp",
        "ebp": "ebp",
        "bp": "bp",
        "bpl": "bpl",
        "rsp": "rsp",
        "esp": "esp",
        "sp": "sp",
        "spl": "spl",
        # *** Instruction Pointer ***
        "rip": "rip",
        "eip": "eip",
        "ip": "ip",
        # *** Segment Registers ***
        "cs": "cs",
        "ds": "ds",
        "es": "es",
        "fs": "fs",
        "gs": "gs",
        "ss": "ss",
        # *** Flags Register ***
        "rflags": "rflags",
        "eflags": "eflags",
        "flags": "flags",
        # *** Control Registers ***
        # NOTE: Ghidra happily supports up to cr15
        "cr0": "cr0",
        "cr1": "cr1",
        "cr2": "cr2",
        "cr3": "cr3",
        "cr4": "cr4",
        "cr8": "cr8",
        # *** Debug Registers ***
        "dr0": "dr0",
        "dr1": "dr1",
        "dr2": "dr2",
        "dr3": "dr3",
        "dr6": "dr6",
        "dr7": "dr7",
        "dr8": "dr8",
        "dr9": "dr9",
        "dr10": "dr10",
        "dr11": "dr11",
        "dr12": "dr12",
        "dr13": "dr13",
        "dr14": "dr14",
        "dr15": "dr15",
        # *** Descriptor Table Registers ***
        "gdtr": "gdtr",
        "idtr": "idtr",
        "ldtr": "ldtr",
        # *** Task Register ***
        "tr": "tr",
        # *** x87 Registers ***
        # TODO: Ghidra appears to support x87, but I have no idea how it works
        "fpr0": None,
        "fpr1": None,
        "fpr2": None,
        "fpr3": None,
        "fpr4": None,
        "fpr5": None,
        "fpr6": None,
        "fpr7": None,
        "fctrl": None,
        "fstat": None,
        "ftag": None,
        "fip": None,
        "fdp": None,
        "fop": None,
        # *** MMX Registers ***
        # TODO: Ghidra appears to support mmx, but I have no idea how it works.
        "mm0": None,
        "mm1": None,
        "mm2": None,
        "mm3": None,
        "mm4": None,
        "mm5": None,
        "mm6": None,
        "mm7": None,
        # SSE/AVX registers
        "ymm0": "ymm0",
        "xmm0": "xmm0",
        "ymm1": "ymm1",
        "xmm1": "xmm1",
        "ymm2": "ymm2",
        "xmm2": "xmm2",
        "ymm3": "ymm3",
        "xmm3": "xmm3",
        "ymm4": "ymm4",
        "xmm4": "xmm4",
        "ymm5": "ymm5",
        "xmm5": "xmm5",
        "ymm6": "ymm6",
        "xmm6": "xmm6",
        "ymm7": "ymm7",
        "xmm7": "xmm7",
        "ymm8": "ymm8",
        "xmm8": "xmm8",
        "ymm9": "ymm9",
        "xmm9": "xmm9",
        "ymm10": "ymm10",
        "xmm10": "xmm10",
        "ymm11": "ymm11",
        "xmm11": "xmm11",
        "ymm12": "ymm12",
        "xmm12": "xmm12",
        "ymm13": "ymm13",
        "xmm13": "xmm13",
        "ymm14": "ymm14",
        "xmm14": "xmm14",
        "ymm15": "ymm15",
        "xmm15": "xmm15",
    }



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import MIPS64ELSysVModel


class MIPS64ELSysVMemcpy(Memcpy, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVMemmove(Memmove, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrcpy(Strcpy, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrncpy(Strncpy, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrcat(Strcat, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrncat(Strncat, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVMemcmp(Memcmp, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrncmp(Strncmp, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrcmp(Strcmp, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrcoll(Strcoll, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrxfrm(Strxfrm, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVMemchr(Memchr, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrchr(Strchr, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrcspn(Strcspn, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrpbrk(Strpbrk, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrrchr(Strrchr, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrspn(Strspn, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrstr(Strstr, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrtok(Strtok, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVMemset(Memset, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrerror(Strerror, MIPS64ELSysVModel):
    pass


class MIPS64ELSysVStrlen(Strlen, MIPS64ELSysVModel):
    pass


__all__ = [
    "MIPS64ELSysVMemcpy",
    "MIPS64ELSysVMemmove",
    "MIPS64ELSysVStrcpy",
    "MIPS64ELSysVStrncpy",
    "MIPS64ELSysVStrcat",
    "MIPS64ELSysVStrncat",
    "MIPS64ELSysVMemcmp",
    "MIPS64ELSysVStrncmp",
    "MIPS64ELSysVStrcmp",
    "MIPS64ELSysVStrcoll",
    "MIPS64ELSysVStrxfrm",
    "MIPS64ELSysVMemchr",
    "MIPS64ELSysVStrchr",
    "MIPS64ELSysVStrcspn",
    "MIPS64ELSysVStrpbrk",
    "MIPS64ELSysVStrrchr",
    "MIPS64ELSysVStrspn",
    "MIPS64ELSysVStrstr",
    "MIPS64ELSysVStrtok",
    "MIPS64ELSysVMemset",
    "MIPS64ELSysVStrerror",
    "MIPS64ELSysVStrlen",
]



from .. import exceptions


class UnicornEmulationError(exceptions.EmulationError):
    def __init__(self, exception: Exception, pc: int, data: dict):
        self.exception = exception
        self.pc = pc
        self.data = data

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.exception}, {hex(self.pc)}, {self.data})"
        )


class AnalysisSetupError(exceptions.AnalysisError):
    """Raised when an analysis run gets into trouble during setup."""


class AnalysisRunError(exceptions.AnalysisError):
    """Raised when something goes wrong during an analysis."""


class AnalysisSignal(exceptions.Error):
    """Raised to signal a non-fatal exception during an analysis."""



from ... import platforms, state
from . import cpu


class XTensa(cpu.CPU):
    """CPU for XTensa, little-endian

    Like RISC-V, which shares its lineage,
    xtensa has a very small core architecture
    with a metric boatload of optional extensions.

    One noteable option is that xtensa uses register windows.
    I'm not putting up with that shit for now.
    """

    def __init__(self):
        super().__init__()
        # *** General Purpose Registers ***
        # a0 is also the default link register, but it doesn't get an alias
        self.a0 = state.Register("a0", 4)
        self.add(self.a0)
        # a1 is also the stack pointer
        self.a1 = state.Register("a1", 4)
        self.add(self.a1)
        self.sp = state.RegisterAlias("sp", self.a1, 4, 0)
        self.add(self.sp)
        self.a2 = state.Register("a2", 4)
        self.add(self.a2)
        self.a3 = state.Register("a3", 4)
        self.add(self.a3)
        self.a4 = state.Register("a4", 4)
        self.add(self.a4)
        self.a5 = state.Register("a5", 4)
        self.add(self.a5)
        self.a6 = state.Register("a6", 4)
        self.add(self.a6)
        self.a7 = state.Register("a7", 4)
        self.add(self.a7)
        self.a8 = state.Register("a8", 4)
        self.add(self.a8)
        self.a9 = state.Register("a9", 4)
        self.add(self.a9)
        self.a10 = state.Register("a10", 4)
        self.add(self.a10)
        self.a11 = state.Register("a11", 4)
        self.add(self.a11)
        self.a12 = state.Register("a12", 4)
        self.add(self.a12)
        self.a13 = state.Register("a13", 4)
        self.add(self.a13)
        self.a14 = state.Register("a14", 4)
        self.add(self.a14)
        self.a15 = state.Register("a15", 4)
        self.add(self.a15)

        # *** Program Counter ***
        self.pc = state.Register("pc", 4)
        self.add(self.pc)

        # *** Shift Amount Register ***
        # This thing is actually 6 bits.
        self.sar = state.Register("sar", 4)
        self.add(self.sar)


class XTensaEL(XTensa):
    platform = platforms.Platform(
        platforms.Architecture.XTENSA, platforms.Byteorder.LITTLE
    )


class XTensaBE(XTensa):
    platform = platforms.Platform(
        platforms.Architecture.XTENSA, platforms.Byteorder.BIG
    )



import sys
import typing

from .. import hinting
from . import analysis

# goal here is to summarize across multiple micro execution if two
# hints map to the same key then they are in same equivalence class so
# we elide away things that will be specific to a run like micro exec
# number, color (actual dyn value), ptr addresses, etc.


def compute_dv_key(hint):
    if type(hint) is hinting.DynamicRegisterValueHint:
        return (
            ("type", "reg"),
            ("pc", hint.pc),
            ("size", hint.size),
            ("use", hint.use),
            ("new", hint.new),
            ("reg_name", hint.reg_name),
        )
    elif type(hint) is hinting.DynamicMemoryValueHint:
        return (
            ("type", "mem"),
            ("pc", hint.pc),
            ("size", hint.size),
            ("use", hint.use),
            ("new", hint.new),
            ("base", hint.base),
            ("index", hint.index),
            ("scale", hint.scale),
            ("offset", hint.offset),
        )
    elif type(hint) is hinting.MemoryUnavailableHint:
        return (
            ("type", "mem_unavail"),
            ("pc", hint.pc),
            ("size", hint.size),
            ("is_read", hint.is_read),
            ("base", hint.base_reg_name),
            ("index", hint.index_reg_name),
            ("scale", hint.scale),
            ("offset", hint.offset),
        )
    elif type(hint) is hinting.EmulationException:
        return (("type", "emu_fail"), ("pc", hint.pc), ("exception", hint.exception))

    else:
        # should never happen
        assert (
            (type(hint) is hinting.DynamicRegisterValueHint)
            or (type(hint) is hinting.DynamicMemoryValueHint)
            or (type(hint) is hinting.MemoryUnavailableHint)
            or (type(hint) is hinting.EmulationException)
        )


class ColorizerSummary(analysis.Analysis):
    name = "colorizer_summary"
    description = "collect and summarize colorizer output across micro executions"
    version = "0.0.1"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hint_list = []
        self.exec_ids = set([])
        self.hinter.register(hinting.DynamicRegisterValueHint, self.collect_hints)
        self.hinter.register(hinting.DynamicMemoryValueHint, self.collect_hints)
        self.hinter.register(hinting.MemoryUnavailableHint, self.collect_hints)
        self.hinter.register(hinting.EmulationException, self.collect_hints)

    def collect_hints(self, hint):
        self.hint_list.append(hint)
        self.exec_ids.add(hint.exec_id)

    def run(self, machine):
        # Establish true colors 1, 2, 3, ...
        #
        # The colors we have in hints are random initial reg or memory
        # values or values computed from those. This means they don't
        # correspond across micro-executions. We need them to do so if
        # we are to summarize.
        #
        truecolors = set([])
        dvh2truecolor = {}
        color2truecolor = {}
        for hint in self.hint_list:
            if hasattr(hint, "color"):
                if hint.new:
                    # This is a hint with a "new" color meaning the
                    # hint time is when the color was first
                    # observed. If we are trying to match up such new
                    # colors across micro executions, we do so by
                    # matching up the dv keys for them at this
                    # time of first observation.
                    dvk = compute_dv_key(hint)
                    dvh = hash(dvk) % ((sys.maxsize + 1) * 2)
                    if dvh not in dvh2truecolor:
                        # we need to establish a color for this dvh
                        next_color = 1 + len(truecolors)
                        truecolors.add(next_color)
                        dvh2truecolor[dvh] = next_color
                    # make sure to map every hint color to its true color
                    color2truecolor[hint.color] = dvh2truecolor[dvh]
                else:
                    assert hint.color in color2truecolor

        # use compute_dv_key to put keys into equiv classes
        # which works for both new and not-new hints
        # start off by collecting set of all such classes
        all_hint_keys = set([])
        hk_exemplar = {}
        for hint in self.hint_list:
            hk = compute_dv_key(hint)
            all_hint_keys.add(hk)
            # keep one exemplar for each equivalence class
            if hk not in hk_exemplar:
                hk_exemplar[hk] = hint

        hint_keys_sorted = sorted(list(all_hint_keys))

        # given the equivalence classes established by `compute_dv_key`, determine
        # which of those were observed in each micro-execution
        hk_observed: typing.Dict[
            int, typing.Set[typing.Tuple[int, bool, str, bool, str, str, str, int, int]]
        ] = {}
        for hint in self.hint_list:
            if hint.exec_id not in hk_observed:
                hk_observed[hint.exec_id] = set([])
                # this hint key was observed in micro execution me
            hk_observed[hint.exec_id].add(compute_dv_key(hint))

        # get counts for hint across micro executions
        hk_c = {}
        for hk in hint_keys_sorted:
            hk_c[hk] = 0
            for exec_id in self.exec_ids:
                for hk2 in hk_observed[exec_id]:
                    if hk == hk2:
                        hk_c[hk] += 1

        for hk in hint_keys_sorted:
            hint = hk_exemplar[hk]
            if type(hint) is hinting.DynamicRegisterValueHint:
                self.hinter.send(
                    hinting.DynamicRegisterValueSummaryHint(
                        pc=hint.pc,
                        reg_name=hint.reg_name,
                        color=color2truecolor[hint.color],
                        size=hint.size,
                        use=hint.use,
                        new=hint.new,
                        count=hk_c[hk],
                        num_micro_executions=len(self.exec_ids),
                        message=hint.message + "-summary",
                    )
                )
            if type(hint) is hinting.DynamicMemoryValueHint:
                self.hinter.send(
                    hinting.DynamicMemoryValueSummaryHint(
                        pc=hint.pc,
                        size=hint.size,
                        base=hint.base,
                        index=hint.index,
                        scale=hint.scale,
                        offset=hint.offset,
                        color=color2truecolor[hint.color],
                        use=hint.use,
                        new=hint.new,
                        message=hint.message + "-summary",
                        count=hk_c[hk],
                        num_micro_executions=len(self.exec_ids),
                    )
                )
            if type(hint) is hinting.MemoryUnavailableHint:
                self.hinter.send(
                    hinting.MemoryUnavailableSummaryHint(
                        is_read=hint.is_read,
                        size=hint.size,
                        base_reg_name=hint.base_reg_name,
                        index_reg_name=hint.index_reg_name,
                        offset=hint.offset,
                        scale=hint.bscale,
                        pc=hint.pc,
                        count=hk_c[hk],
                        num_micro_executions=len(self.exec_ids),
                        message=hint.message + "-summary",
                    )
                )



from ....posix.libgen import Basename, Dirname
from ..systemv import MIPS64SysVModel


class MIPS64SysVBasename(Basename, MIPS64SysVModel):
    pass


class MIPS64SysVDirname(Dirname, MIPS64SysVModel):
    pass


__all__ = [
    "MIPS64SysVBasename",
    "MIPS64SysVDirname",
]



from ....posix.libgen import Basename, Dirname
from ..systemv import ArmHFSysVModel


class ArmHFSysVBasename(Basename, ArmHFSysVModel):
    pass


class ArmHFSysVDirname(Dirname, ArmHFSysVModel):
    pass


__all__ = [
    "ArmHFSysVBasename",
    "ArmHFSysVDirname",
]



import logging

from angr.factory import AngrObjectFactory

from ...exceptions import AnalysisError

log = logging.getLogger(__name__)


class PatchedObjectFactory(AngrObjectFactory):
    """Extension of AngrObjectFactory to allow function overrides

    There are a couple of core functions (blocks)
    which need to get overloaded, but which are not exposed
    to any kind of plugin interface.
    """

    def block(self, *args, **kwargs):
        if "backup_state" in kwargs:
            # Bound block lifting based on our code bounds
            # Angr's Vex lifter will happily run off the edge of memory,
            # interpreting undefined memory as zeroes.
            state = kwargs["backup_state"]
            if state._ip.symbolic:
                raise AnalysisError("Cannot build a block for a symbolic IP")
            ip = state._ip.concrete_value

            # Check if the ip is mapped
            (r, found) = state.scratch.memory_map.find_closest_range(ip)
            if not found:
                # Nope.  No code here.
                log.warn(f"No block mapped at {state._ip}")
                max_size = 0
            else:
                # Yep.  We have an upper bound on our block
                (start, stop) = r
                max_size = stop - ip
                if not state.scratch.bounds.is_empty():
                    # We also have bounds.  Test if we're in those
                    (r, found) = state.scratch.bounds.find_closest_range(ip)
                    if not found:
                        # Nope.  Out of bounds.
                        log.warn(f"{state._ip} is out of bounds")
                        max_size = 0
                    else:
                        # Yep.  Allow anything in bounds and in memory
                        (start, stop) = r
                        max_size = min(max_size, stop - ip)

            if max_size == 0:
                log.warn(f"Empty block at {state._ip}")
            max_size = min(max_size, 4096)
            kwargs["size"] = max_size

        return super().block(*args, **kwargs)



import abc

from ...emulators import Emulator
from ..analysis import Analysis


class AnalysisUnderlay(Analysis):
    """Base class for execution underlays

    Some analyses are orthogonal to
    exactly how their program gets actuated.
    In those cases, the author can write the
    bulk of the analysis in an overlay,
    and pair that overlay with different underlays.
    """

    @property
    def emulator(self) -> Emulator:
        """The emulator to run
        Underlays need the overlay to define the emulator.
        """
        return self._emulator

    @emulator.setter
    def emulator(self, emu: Emulator):
        self._emulator = emu

    @abc.abstractmethod
    def execute(self) -> None:
        """Exercise the emulator"""
        raise NotImplementedError()



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import I386SysVModel


class I386SysVBsdSignal(BsdSignal, I386SysVModel):
    pass


class I386SysVKill(Kill, I386SysVModel):
    pass


class I386SysVKillpg(Killpg, I386SysVModel):
    pass


class I386SysVPthreadKill(PthreadKill, I386SysVModel):
    pass


class I386SysVPthreadSigmask(PthreadSigmask, I386SysVModel):
    pass


class I386SysVSigaction(Sigaction, I386SysVModel):
    pass


class I386SysVSigaddset(Sigaddset, I386SysVModel):
    pass


class I386SysVSigaltstack(Sigaltstack, I386SysVModel):
    pass


class I386SysVSigdelset(Sigdelset, I386SysVModel):
    pass


class I386SysVSigemptyset(Sigemptyset, I386SysVModel):
    pass


class I386SysVSigfillset(Sigfillset, I386SysVModel):
    pass


class I386SysVSighold(Sighold, I386SysVModel):
    pass


class I386SysVSigignore(Sigignore, I386SysVModel):
    pass


class I386SysVSiginterrupt(Siginterrupt, I386SysVModel):
    pass


class I386SysVSigismember(Sigismember, I386SysVModel):
    pass


class I386SysVSigpause(Sigpause, I386SysVModel):
    pass


class I386SysVSigpending(Sigpending, I386SysVModel):
    pass


class I386SysVSigprocmask(Sigprocmask, I386SysVModel):
    pass


class I386SysVSigqueue(Sigqueue, I386SysVModel):
    pass


class I386SysVSigrelse(Sigrelse, I386SysVModel):
    pass


class I386SysVSigset(Sigset, I386SysVModel):
    pass


class I386SysVSigsuspend(Sigsuspend, I386SysVModel):
    pass


class I386SysVSigtimedwait(Sigtimedwait, I386SysVModel):
    pass


class I386SysVSigwait(Sigwait, I386SysVModel):
    pass


class I386SysVSigwaitinfo(Sigwaitinfo, I386SysVModel):
    pass


__all__ = [
    "I386SysVBsdSignal",
    "I386SysVKill",
    "I386SysVKillpg",
    "I386SysVPthreadKill",
    "I386SysVPthreadSigmask",
    "I386SysVSigaction",
    "I386SysVSigaddset",
    "I386SysVSigaltstack",
    "I386SysVSigdelset",
    "I386SysVSigemptyset",
    "I386SysVSigfillset",
    "I386SysVSighold",
    "I386SysVSigignore",
    "I386SysVSiginterrupt",
    "I386SysVSigismember",
    "I386SysVSigpause",
    "I386SysVSigpending",
    "I386SysVSigprocmask",
    "I386SysVSigqueue",
    "I386SysVSigrelse",
    "I386SysVSigset",
    "I386SysVSigsuspend",
    "I386SysVSigtimedwait",
    "I386SysVSigwait",
    "I386SysVSigwaitinfo",
]



import abc
import re
import struct

from ....emulators import Emulator
from ....platforms import Byteorder
from ..cstd import ArgumentType, CStdModel, VariadicContext
from ..filedesc import FileDescriptor


class FormatConversionError(Exception):
    pass


class InputEndedError(Exception):
    pass


class Intake(metaclass=abc.ABCMeta):
    """Abstract class for input stream

    The format scan algorithms need to scan both files
    and strings.  I don't want to write two models.
    """

    def __init__(self):
        self.cursor = 0

    @abc.abstractmethod
    def peek(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def push(self, data: str) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def pop(self) -> None:
        raise NotImplementedError()


class FileIntake(Intake):
    """Input stream backed by a file

    That is, a file modeled by the stdio models.
    """

    def __init__(self, file: FileDescriptor):
        super().__init__()
        self.file = file

    def peek(self) -> str:
        data = self.file.read(1, ungetc=True)
        if len(data) == 0:
            raise InputEndedError()

        self.file.ungetc(data[0])

        if data == b"\0":
            raise InputEndedError()

        return data.decode("utf-8")

    def push(self, data: str) -> None:
        raw = data.encode("utf-8")
        for char in reversed(raw):
            self.file.ungetc(char)

        self.cursor -= len(data)

    def pop(self) -> None:
        self.file.read(1, ungetc=True)
        self.cursor += 1


class StringIntake(Intake):
    """Input stream backed by memory.

    Data is read on-demand from the emulator
    """

    def __init__(self, address: int, emulator: Emulator):
        super().__init__()
        self.address = address
        self.emulator = emulator

    def peek(self) -> str:
        data = self.emulator.read_memory(self.address + self.cursor, 1)
        if data == b"\0":
            raise InputEndedError()
        return data.decode("utf-8")

    def push(self, data: str) -> None:
        self.cursor -= len(data)

    def pop(self) -> None:
        self.cursor += 1


# Conversion specifiers:
#
# Group 1: Zero or one '*' flags
#
# Group 2: Zero or one maximum width specifiers
#
# Group 3: Zero or one length specifiers
#
# Group 4: Conversion ID

# Signed decimal int conversion
#
# Allowed Width Specifiers
# - 'hh': char
# - 'h': short
# - '': int
# - 'l': long
# - 'll', 'L', 'q': long long
# - 'z': ssize_t
# - 'j': intmax_t
# - t: ptrdiff_t
#
# Allowed Conversions:
# - 'd', 'i': signed decimal integer
sint_re = re.compile(r"%([*]?)([0-9]*)(hh|h|l|ll|L|q|z|j|t|)(d|i)")

# Unsigned int conversion
#
# Allowed Width Specifiers
# - 'hh': unsigned char
# - 'h': unsigned short
# - '': unsigned int
# - 'l': unsigned long
# - 'll', 'L', 'q': unsigned long long
# - 'z': size_t
# - 'j': intmax_t
# - t: ptrdiff_t
#
# Allowed Conversions:
# - 'o': unsigned octal integer
# - 'u': unsigned decimal integer
# - 'x', 'X': unsigned hexadecimal integer
uint_re = re.compile(r"%([*]?)([0-9]*)(hh|h|l|ll|L|q|z|j|t|)(o|u|x|X)")

# Floating-point conversion
#
# Allowed Width Specifiers
# - '': float
# - 'l': double
# - 'll', 'L', 'q': long double
#
# Allowed Conversions:
# - 'e', 'E', 'f', 'g', 'G', 'a', 'A': signed floating-point
float_re = re.compile(r"%([*]?)([0-9]*)(l|ll|L|q|)(e|E|f|g|a)")

# Character conversion
#
# Allowed Width Specifiers
# - '': char
# - 'l': wchar_t
#
# Allowed Conversions
# - 'c': Character
char_re = re.compile(r"%([*]?)([0-9]*)(l|)(c)")

# String conversion
#
# Allowed Width Specifiers:
# - '': char
# - 'l': wchar_t
#
# Allowed Conversions:
# - 's': String
string_re = re.compile(r"%([*]?)([0-9]*)(l|)(s)")

# Constrained String Conversion
#
# Allowed Width Specifiers:
# - '': char
# - 'l': wchar_t
#
# Allowed conversions:
# - '[.*]': Constrained string; values between '[' and ']' are allowed characters.
# - '[^.*]': Constrained string; values between '[^' and ']' are forbidden characters.

# Special case to detect '%[]]' and '%[^]]'
# These will satisfy the general regexes, but will match
# on '[]' or '[^]', which isn't correct.
brace_re = re.compile(r"%([*]?)([0-9]*)(l|)\[([\^]?\])\]")
constrained_re = re.compile(r"%([*]?)([0-9]*)(l|)\[([\^]?[\]]?[^\]]*)\]")

# Pointer Conversion
#
# Allowed Width Specifiers:
# - '': void *
#
# Allowed Conversions:
# - 'p': Pointer
pointer_re = re.compile(r"%([*]?)()([0-9]*)(p)")

# Length of current input
#
# Allowed Width Specifiers
# - 'hh': char
# - 'h': short
# - '': int
# - 'l': long
# - 'll', 'L', 'q': long long
# - 'z': ssize_t
# - 'j': intmax_t
# - 't': ptrdiff_t
#
# Allowed Conversions
# - 'n': Store number of characters consumed to this point.
length_re = re.compile(r"%([*]?)(hh|h|l|ll|z|j|t|)([0-9]*)(n)")

# Percent sign
percent_re = re.compile(r"%%")


def handle_sint(
    intake: Intake, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> bool:
    flags = m.group(1)
    maxwidth = -1 if m.group(2) == "" else int(m.group(2))
    length = m.group(3)

    strval = ""

    while True:
        try:
            char = intake.peek()
        except InputEndedError:
            return False

        if char.isspace():
            intake.pop()
        else:
            break

    while maxwidth == -1 or len(strval) < maxwidth:
        try:
            char = intake.peek()
        except InputEndedError:
            break

        if strval == "" and char in ("+", "-"):
            strval += char
            intake.pop()
        elif char in "0123456789":
            strval += char
            intake.pop()
        else:
            break

    if strval == "":
        return False

    if flags == "*":
        intake.push(strval)
        return True

    intval = int(strval)
    if length in ("ll", "L", "q") and intval > (2**63):
        intval &= varargs._long_long_inv_mask & ~varargs._long_long_sign_mask
    elif intval > (2 ** ((varargs.platdef.address_size * 8) - 1)):
        intval &= varargs._long_inv_mask & ~varargs._long_sign_mask

    if intval < 0:
        intval *= -1
        intval = (intval ^ varargs._long_long_inv_mask) + 1

    arg = varargs.get_next_argument(ArgumentType.POINTER, emulator)

    assert isinstance(arg, int)

    if length == "hh":
        intval &= 0xFF
        width = 1
    elif length == "h":
        intval &= 0xFFFF
        width = 2
    elif length == "":
        intval &= varargs._int_inv_mask
        width = 4
    elif length in ("l", "z"):
        intval &= varargs._long_inv_mask
        width = 4 if ArgumentType.LONG in varargs._four_byte_types else 8
    elif length in ("ll", "L", "q"):
        intval &= varargs._long_long_inv_mask
        width = 8
    elif length == "j":
        raise NotImplementedError("Type 'intmax_t' not handled")
    elif length == "t":
        raise NotImplementedError("Type 'ptrdiff_t' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")

    if varargs.platform.byteorder == Byteorder.LITTLE:
        byteval = intval.to_bytes(width, "little")
    else:
        byteval = intval.to_bytes(width, "big")

    emulator.write_memory(arg, byteval)

    return True


def handle_uint(
    intake: Intake, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> bool:
    flags = m.group(1)
    maxwidth = -1 if m.group(2) == "" else int(m.group(2))
    length = m.group(3)
    conversion = m.group(4).lower()

    strval = ""

    while True:
        char = intake.peek()
        if char.isspace():
            intake.pop()
        else:
            break

    while maxwidth == -1 or len(strval) < maxwidth:
        try:
            char = intake.peek()
        except InputEndedError:
            break
        if conversion == "x" and char in ("x", "X") and strval == "0":
            strval += char
            intake.pop()
        elif conversion == "x" and char in "0123456789abcdefABCDEF":
            strval += char
            intake.pop()
        elif conversion == "u" and char in "0123456789":
            strval += char
            intake.pop()
        elif conversion == "o" and char in "01234567":
            strval += char
            intake.pop()
        else:
            break

    if strval == "":
        return False

    if flags == "*":
        intake.push(strval)
        return True

    if conversion == "o":
        intval = int(strval, 8)
    elif conversion == "u":
        intval = int(strval)
    elif conversion == "x":
        intval = int(strval, 16)

    arg = varargs.get_next_argument(ArgumentType.POINTER, emulator)

    assert isinstance(arg, int)

    if length == "hh":
        intval &= 0xFF
        width = 1
    elif length == "h":
        intval &= 0xFFFF
        width = 2
    elif length == "":
        intval &= varargs._int_inv_mask
        width = 4
    elif length in ("l", "z"):
        intval &= varargs._long_inv_mask
        width = 4 if ArgumentType.LONG in varargs._four_byte_types else 8
    elif length in ("ll", "L", "q"):
        intval &= varargs._long_long_inv_mask
        width = 8
    elif length == "j":
        raise NotImplementedError("Type 'intmax_t' not handled")
    elif length == "t":
        raise NotImplementedError("Type 'ptrdiff_t' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")

    if varargs.platform.byteorder == Byteorder.LITTLE:
        byteval = intval.to_bytes(width, "little")
    else:
        byteval = intval.to_bytes(width, "big")
    emulator.write_memory(arg, byteval)

    return True


def handle_float(
    intake: Intake, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> bool:
    flags = m.group(1)
    maxwidth = -1 if m.group(2) == "" else int(m.group(2))
    length = m.group(3)

    found_dot = False
    found_sci = False
    strval = ""
    fixedval = ""

    # NOTE: This will reject hexadecimal floats without a specific error.
    # If _anyone_ finds a legit use of hex floats since 1999, please let me know.
    # I'd be most curious to see.

    while True:
        try:
            char = intake.peek()
        except InputEndedError:
            return False

        if char.isspace():
            intake.pop()
        else:
            break

    while maxwidth == -1 or len(strval) < maxwidth:
        try:
            char = intake.peek()
        except InputEndedError:
            break

        if "infinity".startswith(strval + char):
            strval += char
            intake.pop()
        elif "INFINITY".startswith(strval + char):
            strval += char
            intake.pop()
        elif "nan".startswith(strval + char):
            strval += char
            intake.pop()
        elif "NAN".startswith(strval + char):
            strval += char
            intake.pop()
        elif "-infinity".startswith(strval + char):
            strval += char
            intake.pop()
        elif "-INFINITY".startswith(strval + char):
            strval += char
            intake.pop()
        elif "-nan".startswith(strval + char):
            strval += char
            intake.pop()
        elif "-NAN".startswith(strval + char):
            strval += char
            intake.pop()
        elif (strval == "" or strval[-1] in ("e", "E")) and char in ("+", "-"):
            strval += char
            intake.pop()
        elif not found_dot and not found_sci and char == ".":
            found_dot = True
            strval += char
            intake.pop()
        elif not found_sci and char in ("e", "E"):
            # So, scientific notation is a pain,
            # because an invalid scientific number can be prefixed by a valid fixed-point number.
            # Remember what came before; we'll need to restore back to it.
            found_sci = True
            fixedval = strval
            strval += char
            intake.pop()
        elif char in "0123456789":
            strval += char
            intake.pop()
        else:
            break

    if strval == "":
        return False

    if strval[-1] in ("e", "E"):
        # C accepts scientific notation with no exponent.
        # Python does not.
        strval += "0"

    try:
        floatval = float(strval)
    except ValueError:
        if fixedval == "":
            return False
        try:
            intake.push(strval[len(fixedval) :])
            floatval = float(fixedval)
        except ValueError:
            intake.push(fixedval)
            return False

    if flags == "*":
        return True

    endian = ">" if varargs.platform.byteorder == Byteorder.BIG else "<"

    if length == "":
        byteval = struct.pack(f"{endian}f", floatval)
    elif length == "l":
        byteval = struct.pack(f"{endian}d", floatval)
    elif length == "L":
        raise NotImplementedError("long doubles not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")

    arg = varargs.get_next_argument(ArgumentType.POINTER, emulator)

    assert isinstance(arg, int)

    emulator.write_memory(arg, byteval)

    return True


def handle_char(
    intake: Intake, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> bool:
    flags = m.group(1)
    maxwidth = 1 if m.group(2) == "" else int(m.group(2))
    length = m.group(3)

    if length == "l":
        raise NotImplementedError("wchar_t not handled")
    elif length != "":
        raise FormatConversionError(f"Unknown type specifier {length}")

    strval = ""

    while len(strval) < maxwidth:
        try:
            char = intake.peek()
        except InputEndedError:
            break

        strval += char
        intake.pop()

    if flags == "*":
        return True

    byteval = strval.encode("utf-8")

    arg = varargs.get_next_argument(ArgumentType.POINTER, emulator)

    assert isinstance(arg, int)

    emulator.write_memory(arg, byteval)

    return True


def handle_string(
    intake: Intake, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> bool:
    flags = m.group(1)
    maxwidth = -1 if m.group(2) == "" else int(m.group(2))
    length = m.group(3)

    if length == "l":
        raise NotImplementedError("wchar_t not handled")
    elif length != "":
        raise FormatConversionError(f"Unknown type specifier {length}")

    strval = ""

    while True:
        try:
            char = intake.peek()
        except InputEndedError:
            return False

        if char.isspace():
            intake.pop()
        else:
            break

    while maxwidth == -1 or len(strval) < maxwidth:
        try:
            char = intake.peek()
        except InputEndedError:
            break

        if not char.isspace():
            strval += char
            intake.pop()
        else:
            break

    if flags == "*":
        return True

    byteval = strval.encode("utf-8")
    byteval += b"\0"

    arg = varargs.get_next_argument(ArgumentType.POINTER, emulator)

    assert isinstance(arg, int)

    emulator.write_memory(arg, byteval)

    return True


def handle_constrained(
    intake: Intake, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> bool:
    flags = m.group(1)
    maxwidth = -1 if m.group(2) == "" else int(m.group(2))
    length = m.group(3)
    conversion = m.group(4)

    if length == "l":
        raise NotImplementedError("wchar_t not handled")
    elif length != "":
        raise FormatConversionError(f"Unknown type specifier {length}")

    if conversion[0] == "^":
        invert = True
        conversion = conversion[1:]

    strval = ""
    while maxwidth == -1 or len(strval) < maxwidth:
        try:
            char = intake.peek()
        except InputEndedError:
            break

        if invert and char not in conversion:
            strval += char
            intake.pop()
        elif not invert and char in conversion:
            strval += char
            intake.pop()
        else:
            break

    if flags == "*":
        return True

    byteval = strval.encode("utf-8")
    byteval += b"\0"

    arg = varargs.get_next_argument(ArgumentType.POINTER, emulator)

    assert isinstance(arg, int)

    emulator.write_memory(arg, byteval)

    return True


def handle_pointer(
    intake: Intake, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> bool:
    flags = m.group(1)
    maxwidth = -1 if m.group(2) == "" else int(m.group(2))

    strval = ""

    while True:
        try:
            char = intake.peek()
        except InputEndedError:
            return False

        if char.isspace():
            intake.pop()
        else:
            break

    while maxwidth == -1 or len(strval) < maxwidth:
        try:
            char = intake.peek()
        except InputEndedError:
            break

        if (
            "(nil)".startswith(strval + char)
            or "(NIL)".startswith(strval + char)
            or "0x".startswith(strval + char)
        ):
            strval += char
            intake.pop()
        elif char in "0123456789abcdefABCDEF":
            strval += char
            intake.pop()
        else:
            break

    if strval == "":
        return False

    if flags == "*":
        intake.push(strval)
        return True

    if strval in ("(nil)", "(NIL)"):
        intval = 0
    else:
        try:
            intval = int(strval, 16)
        except ValueError:
            intake.push(strval)
            return False

    arg = varargs.get_next_argument(ArgumentType.POINTER, emulator)

    assert isinstance(arg, int)

    if varargs.platform.byteorder == Byteorder.BIG:
        byteval = intval.to_bytes(varargs.platdef.address_size, "big")
    else:
        byteval = intval.to_bytes(varargs.platdef.address_size, "little")

    emulator.write_memory(arg, byteval)

    return True


def handle_length(
    intake: Intake, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> bool:
    length = m.group(3)

    intval = intake.cursor

    if length == "hh":
        intval &= 0xFF
        width = 1
    elif length == "h":
        intval &= 0xFFFF
        width = 2
    elif length == "":
        intval &= varargs._int_inv_mask
        width = 4
    elif length in ("l", "z"):
        intval &= varargs._long_inv_mask
        width = 4 if ArgumentType.LONG in varargs._four_byte_types else 8
    elif length in ("ll", "L", "q"):
        intval &= varargs._long_long_inv_mask
        width = 8
    elif length == "j":
        raise NotImplementedError("Type 'intmax_t' not handled")
    elif length == "t":
        raise NotImplementedError("Type 'ptrdiff_t' not handled")
    else:
        raise FormatConversionError(f"Unknown type specifier {length}")

    arg = varargs.get_next_argument(ArgumentType.POINTER, emulator)

    assert isinstance(arg, int)

    if varargs.platform.byteorder == Byteorder.BIG:
        byteval = intval.to_bytes(width, "big")
    else:
        byteval = intval.to_bytes(width, "little")

    emulator.write_memory(arg, byteval)

    return True


def handle_percent(
    intake: Intake, varargs: VariadicContext, m: re.Match, emulator: Emulator
) -> bool:
    return True


def handle_scanf_format(
    model: CStdModel, intake: Intake, fmt: str, emulator: Emulator
) -> int:
    """Process a scanf format string

    This is not an entirely complete model.
    The following features are missing:

    - Decoding for hexadecimal floats
    - intmax_t integers
    - ptrdiff_t integers

    Arguments:
        model: The parent API model
        intake: The intake structure from which to fetch data
        emulator: The emulator to run against

    Returns:
        The number of converted arguments, or -1, as per the scanf family.
    """
    varargs = model.get_varargs()

    orig_fmt = fmt

    handlers = [
        (sint_re, handle_sint),
        (uint_re, handle_uint),
        (float_re, handle_float),
        (char_re, handle_char),
        (string_re, handle_string),
        (brace_re, handle_constrained),
        (constrained_re, handle_constrained),
        (pointer_re, handle_pointer),
        (length_re, handle_length),
        (percent_re, handle_percent),
    ]

    converted = 0
    done = False
    while len(fmt) > 0:
        if fmt[0] == "%":
            # Conversion pattern.  See if we can match
            matched = False
            for regex, handler in handlers:
                m = regex.match(fmt)
                if m is not None:
                    try:
                        if not handler(intake, varargs, m, emulator):
                            done = True
                        elif m.group(1) != "*" and m.group(4)[-1] != "n":
                            converted += 1
                    except FormatConversionError as e:
                        print(f"Bad format conversion: {e.args[0]}")
                        print(orig_fmt)
                        print(" " * (len(orig_fmt) - len(fmt)) + "^")
                        raise e
                    except Exception as e:
                        print(f"Exception processing conversion: {type(e)}: {e}")
                        print(orig_fmt)
                        print(" " * (len(orig_fmt) - len(fmt)) + "^")
                        raise e
                    fmt = fmt[len(m.group(0)) :]
                    matched = True
                    break

            if not matched:
                print("Bad format conversion: Unmatched conversion")
                print(orig_fmt)
                print(" " * (len(orig_fmt) - len(fmt)) + "^")
                raise Exception("Bad format conversion")

            if done:
                break

        elif fmt[0].isspace():
            # Whitespace.  Consume zero or more whitespace characters
            while True:
                char = intake.peek()
                if char.isspace():
                    print("Space")
                    intake.pop()
                else:
                    break
            fmt = fmt[1:]
        else:
            # Normal character.  Consume one character that's an exact match
            char = intake.peek()
            if char == fmt[0]:
                intake.pop()
            else:
                return -1

    return converted


__all__ = ["FileIntake", "StringIntake", "handle_scanf_format"]



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import MIPS64SysVModel


class MIPS64SysVBsdSignal(BsdSignal, MIPS64SysVModel):
    pass


class MIPS64SysVKill(Kill, MIPS64SysVModel):
    pass


class MIPS64SysVKillpg(Killpg, MIPS64SysVModel):
    pass


class MIPS64SysVPthreadKill(PthreadKill, MIPS64SysVModel):
    pass


class MIPS64SysVPthreadSigmask(PthreadSigmask, MIPS64SysVModel):
    pass


class MIPS64SysVSigaction(Sigaction, MIPS64SysVModel):
    pass


class MIPS64SysVSigaddset(Sigaddset, MIPS64SysVModel):
    pass


class MIPS64SysVSigaltstack(Sigaltstack, MIPS64SysVModel):
    pass


class MIPS64SysVSigdelset(Sigdelset, MIPS64SysVModel):
    pass


class MIPS64SysVSigemptyset(Sigemptyset, MIPS64SysVModel):
    pass


class MIPS64SysVSigfillset(Sigfillset, MIPS64SysVModel):
    pass


class MIPS64SysVSighold(Sighold, MIPS64SysVModel):
    pass


class MIPS64SysVSigignore(Sigignore, MIPS64SysVModel):
    pass


class MIPS64SysVSiginterrupt(Siginterrupt, MIPS64SysVModel):
    pass


class MIPS64SysVSigismember(Sigismember, MIPS64SysVModel):
    pass


class MIPS64SysVSigpause(Sigpause, MIPS64SysVModel):
    pass


class MIPS64SysVSigpending(Sigpending, MIPS64SysVModel):
    pass


class MIPS64SysVSigprocmask(Sigprocmask, MIPS64SysVModel):
    pass


class MIPS64SysVSigqueue(Sigqueue, MIPS64SysVModel):
    pass


class MIPS64SysVSigrelse(Sigrelse, MIPS64SysVModel):
    pass


class MIPS64SysVSigset(Sigset, MIPS64SysVModel):
    pass


class MIPS64SysVSigsuspend(Sigsuspend, MIPS64SysVModel):
    pass


class MIPS64SysVSigtimedwait(Sigtimedwait, MIPS64SysVModel):
    pass


class MIPS64SysVSigwait(Sigwait, MIPS64SysVModel):
    pass


class MIPS64SysVSigwaitinfo(Sigwaitinfo, MIPS64SysVModel):
    pass


__all__ = [
    "MIPS64SysVBsdSignal",
    "MIPS64SysVKill",
    "MIPS64SysVKillpg",
    "MIPS64SysVPthreadKill",
    "MIPS64SysVPthreadSigmask",
    "MIPS64SysVSigaction",
    "MIPS64SysVSigaddset",
    "MIPS64SysVSigaltstack",
    "MIPS64SysVSigdelset",
    "MIPS64SysVSigemptyset",
    "MIPS64SysVSigfillset",
    "MIPS64SysVSighold",
    "MIPS64SysVSigignore",
    "MIPS64SysVSiginterrupt",
    "MIPS64SysVSigismember",
    "MIPS64SysVSigpause",
    "MIPS64SysVSigpending",
    "MIPS64SysVSigprocmask",
    "MIPS64SysVSigqueue",
    "MIPS64SysVSigrelse",
    "MIPS64SysVSigset",
    "MIPS64SysVSigsuspend",
    "MIPS64SysVSigtimedwait",
    "MIPS64SysVSigwait",
    "MIPS64SysVSigwaitinfo",
]



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import AMD64SysVModel


class AMD64SysVAbs(Abs, AMD64SysVModel):
    pass


class AMD64SysVLAbs(LAbs, AMD64SysVModel):
    pass


class AMD64SysVLLAbs(LLAbs, AMD64SysVModel):
    pass


class AMD64SysVAtof(Atof, AMD64SysVModel):
    pass


class AMD64SysVAtoi(Atoi, AMD64SysVModel):
    pass


class AMD64SysVAtol(Atol, AMD64SysVModel):
    pass


class AMD64SysVAtoll(Atoll, AMD64SysVModel):
    pass


class AMD64SysVCalloc(Calloc, AMD64SysVModel):
    pass


class AMD64SysVDiv(Div, AMD64SysVModel):
    pass


class AMD64SysVLDiv(LDiv, AMD64SysVModel):
    pass


class AMD64SysVLLDiv(LLDiv, AMD64SysVModel):
    pass


class AMD64SysVExit(Exit, AMD64SysVModel):
    pass


class AMD64SysVFree(Free, AMD64SysVModel):
    pass


class AMD64SysVMalloc(Malloc, AMD64SysVModel):
    pass


class AMD64SysVQSort(QSort, AMD64SysVModel):
    pass


class AMD64SysVRand(Rand, AMD64SysVModel):
    pass


class AMD64SysVRealloc(Realloc, AMD64SysVModel):
    pass


class AMD64SysVSrand(Srand, AMD64SysVModel):
    pass


class AMD64SysVAbort(Abort, AMD64SysVModel):
    pass


class AMD64SysVAtexit(Atexit, AMD64SysVModel):
    pass


class AMD64SysVGetenv(Getenv, AMD64SysVModel):
    pass


class AMD64SysVMblen(Mblen, AMD64SysVModel):
    pass


class AMD64SysVMbstowcs(Mbstowcs, AMD64SysVModel):
    pass


class AMD64SysVMbtowc(Mbtowc, AMD64SysVModel):
    pass


class AMD64SysVSystem(System, AMD64SysVModel):
    pass


class AMD64SysVWcstombs(Wcstombs, AMD64SysVModel):
    pass


class AMD64SysVWctomb(Wctomb, AMD64SysVModel):
    pass


__all__ = [
    "AMD64SysVAbs",
    "AMD64SysVLAbs",
    "AMD64SysVLLAbs",
    "AMD64SysVAtof",
    "AMD64SysVAtoi",
    "AMD64SysVAtol",
    "AMD64SysVAtoll",
    "AMD64SysVCalloc",
    "AMD64SysVDiv",
    "AMD64SysVLDiv",
    "AMD64SysVLLDiv",
    "AMD64SysVExit",
    "AMD64SysVFree",
    "AMD64SysVMalloc",
    "AMD64SysVQSort",
    "AMD64SysVRand",
    "AMD64SysVRealloc",
    "AMD64SysVSrand",
    "AMD64SysVAbort",
    "AMD64SysVAtexit",
    "AMD64SysVGetenv",
    "AMD64SysVMblen",
    "AMD64SysVMbstowcs",
    "AMD64SysVMbtowc",
    "AMD64SysVSystem",
    "AMD64SysVWcstombs",
    "AMD64SysVWctomb",
]



import logging

from .... import emulators

logger = logging.getLogger(__name__)

# Maximum string length.
# Used to terminate unbounded string operations
MAX_STRLEN = 0x10000

# Common memory manipulation operations used by a lot of library functions.


def _emu_strnlen(emulator: emulators.Emulator, addr: int, n: int) -> int:
    sl = 0
    while sl <= n:
        b_opt = emulator.read_memory(addr + sl, 1)
        if b_opt is not None:
            b = b_opt[0]
            if b == 0:
                break
            sl += 1
        else:
            assert b_opt is not None
    return sl


def _emu_strlen(emulator: emulators.Emulator, addr: int) -> int:
    return _emu_strnlen(emulator, addr, MAX_STRLEN)


def _emu_memcpy(emulator: emulators.Emulator, dst: int, src: int, n: int) -> None:
    src_bytes = emulator.read_memory(src, n)
    emulator.write_memory(dst, src_bytes)


def _emu_strncpy(emulator: emulators.Emulator, dst: int, src: int, n: int) -> None:
    # strncpy is slightly different from strcpy and memcpy;
    # it zero-fills any unused space in the buffer.
    actual = n
    src_len = _emu_strlen(emulator, src) + 1
    if src_len < n:
        actual = src_len

    data = emulator.read_memory(src, actual)
    data += b"\0" * (n - actual)

    print(f"Writing {n} bytes to {hex(dst)}")
    emulator.write_memory(dst, data)


def _emu_memcmp(emulator: emulators.Emulator, ptr1: int, ptr2: int, n: int) -> int:
    for i in range(0, n):
        char1 = emulator.read_memory(ptr1 + i, 1)[0]
        char2 = emulator.read_memory(ptr2 + i, 1)[0]
        if char1 != char2:
            return char1 - char2
    return 0


def _emu_strncmp(emulator: emulators.Emulator, ptr1: int, ptr2: int, n: int) -> int:
    for i in range(0, n):
        char1 = emulator.read_memory(ptr1 + i, 1)[0]
        char2 = emulator.read_memory(ptr2 + i, 1)[0]
        if char1 == 0 or char2 == 0:
            return char1 - char2
        elif char1 != char2:
            return char1 - char2
    return 0


def _emu_strncat(emulator: emulators.Emulator, dst: int, src: int, n: int) -> None:
    if n == 0:
        return
    if emulator.read_memory(src, 1) is None:
        logger.debug("MEM not available in strncpy read @ {src:x}")
    elif emulator.read_memory(dst, 1) is None:
        logger.debug("MEM not available in strncpy write @ {dst:x}")
    else:
        # at least a byte is available at both src and dst
        ld = _emu_strnlen(emulator, dst, MAX_STRLEN)
        ls = _emu_strnlen(emulator, src, MAX_STRLEN)
        lsn = min(ls, n)
        b_opt = emulator.read_memory(src, lsn)
        if b_opt is not None:
            src_bytes = b_opt + b"\0"
            emulator.write_memory(dst + ld, src_bytes)
        else:
            assert b_opt is not None



from __future__ import annotations

import abc
import typing

from ..... import platforms, utils
from .....exceptions import ConfigurationError
from .. import elf
from ..structs import ElfRela


class ElfRelocator:
    """Platform-specific ELF relocator

    Each platform defines its own set of relocation types,
    defining how symbol values can update an image at link-time.


    """

    @property
    @abc.abstractmethod
    def arch(self) -> platforms.Architecture:
        """The architecture ID"""
        raise NotImplementedError("This is an abstract method")

    @property
    @abc.abstractmethod
    def byteorder(self) -> platforms.Byteorder:
        """The byte order"""
        raise NotImplementedError("This is an abstract method.")

    @abc.abstractmethod
    def _compute_value(self, rela: ElfRela) -> bytes:
        raise NotImplementedError("This is an abstract method.")

    def relocate(self, elf: elf.ElfExecutable, rela: ElfRela) -> None:
        val = self._compute_value(rela)
        for off, seg in elf.items():
            start = off + elf.address
            stop = start + seg.get_size()

            if rela.offset >= start and rela.offset < stop:
                start = rela.offset - start
                end = start + len(val)

                contents = typing.cast(bytes, seg.get_content())
                contents = contents[0:start] + val + contents[end:]
                seg.set_content(contents)
                return
        raise ConfigurationError(
            f"No segment in ELF loaded at {hex(elf.address)} covers address {hex(rela.offset)}"
        )

    @classmethod
    def for_platform(cls, platform: platforms.Platform):
        try:
            return utils.find_subclass(
                cls,
                lambda x: x.arch == platform.architecture
                and x.byteorder == platform.byteorder,
            )
        except:
            raise ValueError(f"No relocator for {platform}")



from ....posix.libgen import Basename, Dirname
from ..systemv import AMD64SysVModel


class AMD64SysVBasename(Basename, AMD64SysVModel):
    pass


class AMD64SysVDirname(Dirname, AMD64SysVModel):
    pass


__all__ = [
    "AMD64SysVBasename",
    "AMD64SysVDirname",
]



import struct

from ..... import emulators, platforms
from ...cstd import ArgumentType, CStdModel


class RiscV64SysVModel(CStdModel):
    """Base class for C models using the AArch64 System V ABI"""

    platform = platforms.Platform(
        platforms.Architecture.RISCV64, platforms.Byteorder.LITTLE
    )
    abi = platforms.ABI.SYSTEMV

    _int_sign_mask = 0x80000000
    _int_inv_mask = 0xFFFFFFFF
    _long_sign_mask = 0x8000000000000000
    _long_inv_mask = 0xFFFFFFFFFFFFFFFF
    _long_long_sign_mask = 0x8000000000000000
    _long_long_inv_mask = 0xFFFFFFFFFFFFFFFF

    _four_byte_types = {ArgumentType.INT, ArgumentType.UINT}

    _eight_byte_types = {
        ArgumentType.LONG,
        ArgumentType.ULONG,
        ArgumentType.LONGLONG,
        ArgumentType.ULONGLONG,
        ArgumentType.SIZE_T,
        ArgumentType.SSIZE_T,
        ArgumentType.POINTER,
    }

    _four_byte_arg_regs = ["a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7"]

    _eight_byte_arg_regs = [
        "a0",
        "a1",
        "a2",
        "a3",
        "a4",
        "a5",
        "a6",
        "a7",
    ]

    _soft_float = False
    _variadic_soft_float = True
    _floats_are_doubles = False
    _float_arg_regs = ["fa0", "fa1", "fa2", "fa3", "fa4", "fa5", "fa6"]

    _double_arg_regs = ["fa0", "fa1", "fa2", "fa3", "fa4", "fa5", "fa6"]

    _init_stack_offset = 0
    _align_stack = False
    _eight_byte_reg_size = 1
    _double_reg_size = 1
    _four_byte_stack_size = 8
    _eight_byte_stack_size = 8
    _float_stack_size = 8
    _double_stack_size = 8

    def _return_4_byte(self, emulator: emulators.Emulator, val: int) -> None:
        """Return a four-byte type"""
        val &= self._int_inv_mask
        # riscv64 sign-extends 4-byte ints to fill the register.
        if val & self._int_sign_mask != 0:
            val |= self._int_signext_mask
        emulator.write_register("a0", val)

    def _return_8_byte(self, emulator: emulators.Emulator, val: int) -> None:
        """Return an eight-byte type"""
        emulator.write_register("a0", val)

    def _return_float(self, emulator: emulators.Emulator, val: float) -> None:
        """Return a float"""
        data = struct.pack("<f", val)
        intval = int.from_bytes(data, "little")
        emulator.write_register("fa0", intval)

    def _return_double(self, emulator: emulators.Emulator, val: float) -> None:
        """Return a double"""
        data = struct.pack("<d", val)
        intval = int.from_bytes(data, "little")
        emulator.write_register("fa0", intval)



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import MIPSSysVModel


class MIPSSysVAbs(Abs, MIPSSysVModel):
    pass


class MIPSSysVLAbs(LAbs, MIPSSysVModel):
    pass


class MIPSSysVLLAbs(LLAbs, MIPSSysVModel):
    pass


class MIPSSysVAtof(Atof, MIPSSysVModel):
    pass


class MIPSSysVAtoi(Atoi, MIPSSysVModel):
    pass


class MIPSSysVAtol(Atol, MIPSSysVModel):
    pass


class MIPSSysVAtoll(Atoll, MIPSSysVModel):
    pass


class MIPSSysVCalloc(Calloc, MIPSSysVModel):
    pass


class MIPSSysVDiv(Div, MIPSSysVModel):
    pass


class MIPSSysVLDiv(LDiv, MIPSSysVModel):
    pass


class MIPSSysVLLDiv(LLDiv, MIPSSysVModel):
    pass


class MIPSSysVExit(Exit, MIPSSysVModel):
    pass


class MIPSSysVFree(Free, MIPSSysVModel):
    pass


class MIPSSysVMalloc(Malloc, MIPSSysVModel):
    pass


class MIPSSysVQSort(QSort, MIPSSysVModel):
    pass


class MIPSSysVRand(Rand, MIPSSysVModel):
    pass


class MIPSSysVRealloc(Realloc, MIPSSysVModel):
    pass


class MIPSSysVSrand(Srand, MIPSSysVModel):
    pass


class MIPSSysVAbort(Abort, MIPSSysVModel):
    pass


class MIPSSysVAtexit(Atexit, MIPSSysVModel):
    pass


class MIPSSysVGetenv(Getenv, MIPSSysVModel):
    pass


class MIPSSysVMblen(Mblen, MIPSSysVModel):
    pass


class MIPSSysVMbstowcs(Mbstowcs, MIPSSysVModel):
    pass


class MIPSSysVMbtowc(Mbtowc, MIPSSysVModel):
    pass


class MIPSSysVSystem(System, MIPSSysVModel):
    pass


class MIPSSysVWcstombs(Wcstombs, MIPSSysVModel):
    pass


class MIPSSysVWctomb(Wctomb, MIPSSysVModel):
    pass


__all__ = [
    "MIPSSysVAbs",
    "MIPSSysVLAbs",
    "MIPSSysVLLAbs",
    "MIPSSysVAtof",
    "MIPSSysVAtoi",
    "MIPSSysVAtol",
    "MIPSSysVAtoll",
    "MIPSSysVCalloc",
    "MIPSSysVDiv",
    "MIPSSysVLDiv",
    "MIPSSysVLLDiv",
    "MIPSSysVExit",
    "MIPSSysVFree",
    "MIPSSysVMalloc",
    "MIPSSysVQSort",
    "MIPSSysVRand",
    "MIPSSysVRealloc",
    "MIPSSysVSrand",
    "MIPSSysVAbort",
    "MIPSSysVAtexit",
    "MIPSSysVGetenv",
    "MIPSSysVMblen",
    "MIPSSysVMbstowcs",
    "MIPSSysVMbtowc",
    "MIPSSysVSystem",
    "MIPSSysVWcstombs",
    "MIPSSysVWctomb",
]



import logging
import typing

import lief

from ..emulators import Emulator
from ..exceptions import ConfigurationError
from .state import Code, Memory

log = logging.getLogger(__name__)

# Prorgam header types
PT_NULL = 0  # Empty/unused program header
PT_LOAD = 1  # Describes loadable program segment
PT_DYNAMIC = 2  # Points to dynamic linking metadata
PT_INTERP = 3  # Points to program interpreter
PT_NOTE = 4  # Points to auxiliary information
PT_SHLIB = 5  # Reserved value; I think it's unused
PT_PHDR = 6  # Points to program header table
PT_TLS = 7  # Indicates need for thread-local storage
PT_LOOS = 0x60000000  # Start of OS-specific types
PT_GNU_EH_FRAME = 0x6474E550  # GNU-specific: Points to exception handler segment
PT_GNU_STACK = 0x6474E551  # GNU-specific: Describes stack permissions
PT_GNU_RELRO = 0x6474E552  # GNU-specific: Describes read-only after relocation segment
PT_GNU_PROPERTY = 0x6474E553  # GNU-specific: Points to GNU property
PT_HIOS = 0x6FFFFFFF  # End of OS-specific types
PT_LOPROC = 0x70000000  # Start of processor-specific types
PT_HIPROC = 0x7FFFFFFF  # End of processor-specific types

# Program header flags
PF_X = 0x1  # Segment is executable
PF_W = 0x2  # Segment is writable
PF_R = 0x4  # Segment is readable


class ELFImage(Code):
    default_base = 0x400000
    page_size = 0x1000

    def __init__(
        self,
        image: bytes,
        format: typing.Optional[str] = None,
        arch: typing.Optional[str] = None,
        mode: typing.Optional[str] = None,
        base: typing.Optional[int] = None,
        entry: typing.Optional[int] = None,
        bounds: typing.Optional[typing.Iterable[range]] = None,
    ):
        super().__init__(
            image,
            base=0,
            format=format,
            arch=arch,
            mode=mode,
            entry=None,
            bounds=bounds,
        )
        self.user_base: typing.Optional[int] = base
        self.file_base: int = 0
        self.user_entry: typing.Optional[int] = entry
        self.file_entry: int = 0

        self.code_segments: typing.List[Code] = list()
        self.data_segments: typing.List[Memory] = list()

        self.load_elf()

    def determine_base(self):
        """Determine base address to load ELF

        There are two possible sources for a base address.
        ELF files can specify a base address by setting the address
        of their first loaded segment to non-zero.
        Otherwise, a user can request a specific base address.

        If both the user and the file specify an address, it's an error.
        If only one specifies a base address, use it.
        If neither specify one, halucinate a value.

        Raises:
            ConfigurationError: If file_base is non-zero, and user_base is not None
        """
        if self.user_base is None:
            # No base address requested
            if self.file_base == 0:
                # Progam file does not need a specific base address
                # Use a default
                # FIXME: Using a fixed value will not be valid if we load multiple files
                log.info(
                    f"No base address requested, and ELF is PIC.  Using {hex(self.default_base)}"
                )
                self.base = self.default_base
            else:
                # Program file needs a specific base address
                # Use the specified address
                self.base = self.file_base
        else:
            # Base address requested
            if self.file_base == 0:
                # Program file does not need a specific base address
                # Use the requested address
                self.user_base
            elif self.file_base == self.user_base:
                # User and file request the same base address.
                # We are okay with this.
                self.user_base
            else:
                # Program file needs a specific base address
                # Not possible to rebase
                log.error(
                    f"Requested base address {hex(self.user_base)}, but program needs {hex(self.file_base)}"
                )
                raise ConfigurationError("Contradictory base addresses.")

    def determine_entry(self):
        """Determine entrypoint address to use

        This performs two jobs.
        One is to determine which entrypoint to use;
        it can be specified by the ELF header, or by the user.

        The second job is to rebase that address
        according to the base address.
        An entrypoint from the file is relative to the file's base address.
        An entrypoint from the user is assumed relative to the user's base address.
        If the user does not specify a base address, their entrypoint
        should be relative to the start of the file.

        Raises:
            ConfigurationError: If neither user nor file specify an entrypoint

        """
        if self.user_entry is None:
            # No entrypoint requested.
            # Get entry from the file
            if self.file_entry == 0:
                # No one specified an entrypoint.
                # No entrypoint will be set for this file.
                self.entry = None
            # file_entry is relative to file_base; rebase relative to base
            self.entry = self.file_entry - self.file_base + self.base
        else:
            # Entrypoint requested.
            if self.user_base is None:
                user_base = 0
            else:
                user_base = self.user_base
            # user_entry is relative to user_base, or zero if none requested
            # rebase relative to base
            self.entry = self.user_entry - user_base + self.base

    def rebase_user(self, addr: int):
        """Helper function for rebasing user-relative addresses"""
        if self.user_base is None:
            old_base = 0
        else:
            old_base = self.user_base
        if self.base is None:
            base = 0
        else:
            base = self.base
        return addr - old_base + base

    def rebase_file(self, addr: int):
        """Helper function for rebasing file-relative addresses"""
        if self.base is None:
            base = 0
        else:
            base = self.base
        return addr - self.file_base + base

    def page_align(self, x: int, up: bool = True):
        """Align an address to a page boundary

        Arguments:
            x (int): Address to align
            up (bool): If true, round up.  If false, round down

        Return:
            (int): Aligned version of the address
        """
        if up:
            x += self.page_size - 1
        return (x // self.page_size) * self.page_size

    def load_elf(self):
        """Load an ELF file into a SmallWorld machine state object.

        This parses the ELF header and the program headers,
        does minimal validation, and then maps the loadable segments into memory.
        It also reports RE-relevant features of the binary as hints.

        This function assumes that the program should be loaded
        using the architecture and mode from the CPU object.

        The user can specify a base address and entrypoint.
        If not specified, this function defaults to the addresses specified in the ELF.

        If both user and ELF specify a base address,
        this function will raise an exception.

        If netither user nor ELF specify a base address,
        a default will be used.

        If both user and ELF specify an entrypoint,
        the user's entrypoint takes precedence.

        If neither user nor elf specify an entrypoint,
        no entrypoint will be set for this file.

        Raises:
            ConfigurationError: If the ELF is invalid, or user-provided data conflicts with ELF.
        """
        # Use lief to check if this is an ELF
        # NOTE: for some reason, this takes list[int], not bytes
        if not lief.is_elf(list(self.image)):
            log.error("File is not an elf.")
            raise ConfigurationError("Input is not an elf")

        # Use lief to parse the ELF
        # NOTE: for some reason, this takes list[int], not bytes
        elf = lief.ELF.parse(list(self.image))
        if elf is None:
            raise ConfigurationError("Failed parsing input")

        ehdr = elf.header
        if ehdr is None:
            raise ConfigurationError("Failed extracting Elf header")

        # TODO: Check machine compatibility?
        # - ei_class
        # - ei_data
        # - ei_osabi (some ABIs)
        # - e_machine
        # - e_flags (some ABIs)

        # Figure out if this file is loadable.
        # Easiest way to tell is if there are program headers
        if ehdr.program_header_offset == 0:
            # NULL phoff means no program headers.
            # This file is not loadable; time to use another ELF loader.
            log.error("No program headers; file is not loadable")
            raise ConfigurationError("File not loadable")
        if ehdr.program_header_offset >= len(self.image):
            log.error(
                f"Invalid program header offset: {hex(ehdr.program_header_offset)}"
            )
            raise ConfigurationError("Invalid program header offset")

        # Determine the file base address,
        # entrypoint, and exit points
        self.file_base = elf.imagebase
        self.file_entry = elf.entrypoint
        self.determine_base()
        self.determine_entry()
        self.exits = list(map(lambda x: self.rebase_user(x), self.exits))

        for phdr in elf.segments:
            log.debug(f"{phdr}")
            if phdr.type == PT_LOAD:
                # Loadable segment
                # Map its data into memory
                self.map_segment(phdr)
            elif phdr.type == PT_DYNAMIC:
                # Dynamic linking metadata.
                # This ELF needs dynamic linking
                log.info("Program includes dynamic linking metadata")
            elif phdr.type == PT_INTERP:
                # Program interpreter
                # This completely changes how program loading works.
                # Whether you care is a different matter.
                interp = self.image[
                    phdr.file_offset : phdr.file_offset + phdr.physical_size
                ]
                log.info(f"Program specifies interpreter {interp!r}")
            elif phdr.type == PT_NOTE:
                # Auxiliary information
                # Possibly useful for comparing machine/OS type.
                pass
            elif phdr.type == PT_PHDR:
                # Program header self-reference
                # Useful for the dynamic linker, but not for us
                pass
            elif phdr.type == PT_TLS:
                # TLS Segment
                # Your analysis is about to get nasty :(
                log.info("Program includes thread-local storage")
            elif phdr.type == PT_GNU_EH_FRAME:
                # Exception handler frame.
                # GCC puts one of these in everything.  Do we care?
                pass
            elif phdr.type == PT_GNU_STACK:
                # Stack executability
                # If this is missing, assume executable stack
                log.info("Program specifies stack permissions")
            elif phdr.type == PT_GNU_RELRO:
                # Read-only after relocation
                # Only the dynamic linker should write this data.
                log.info("Program specifies RELRO data")
            elif phdr.type == PT_GNU_PROPERTY:
                # GNU property segment
                # Contains extra metadata which I'm not sure anything uses
                pass
            elif phdr.type >= PT_LOOS and phdr.type <= PT_HIOS:
                # Unknown OS-specific program header
                # Either this is a weird ISA that extends the generic GNU ABI,
                # or this isn't a Linux ELF.
                log.warn(f"Unknown OS-specific program header: {phdr.type:08x}")
            elif phdr.type >= PT_LOPROC and phdr.type <= PT_HIPROC:
                # Unknown machine-specific program header
                # This is probably a non-Intel ISA.
                # Most of these are harmless, serving to tell the RTLD
                # where to find machine-specific metadata
                log.warn(f"Unknown machine-specific program header: {phdr.type:08x}")
            else:
                # Unknown program header outside the allowed custom ranges
                log.warn(f"Invalid program header: {phdr.type:08x}")

    def map_segment(self, phdr):
        """Map a segment into a SmallWorld machine state object

        This computes the actual mapping boundaries from a LOAD segment,
        and then maps it into Smallworld.

        Executable segments are mapped as Code objects.
        Other segments are mapped as Memory.

        Arguments:
            phdr: Program header object to load
        """

        # Compute segment boundaries
        seg_start = self.page_align(phdr.file_offset, up=False)
        seg_end = self.page_align(phdr.file_offset + phdr.physical_size)
        seg_addr = self.page_align(self.rebase_file(phdr.virtual_address), up=False)
        seg_size = self.page_align(phdr.virtual_size + (phdr.file_offset - seg_start))

        log.debug(f"{phdr.physical_size:012x}")

        log.debug("Mapping: ")
        log.debug(f"    f: [ {seg_start:012x} -> {seg_end:012x} ]")
        log.debug(f"    m: [ {seg_addr:012x} -> {seg_addr + seg_size:012x} ]")

        # Extract segment data, and zero-fill out to seg_size
        seg_data = self.image[seg_start:seg_end]
        if len(seg_data) < seg_size:
            seg_data += b"\x00" * (seg_size - len(seg_data))
        elif len(seg_data) != seg_size:
            # Virtual size should always be greater or equal to file size.
            # If not, something's wrong.
            raise ConfigurationError(
                f"Expected segment of size {seg_size}, but got {len(seg_data)}"
            )

        if (phdr.flags & PF_X) != 0:
            # If this is an executable segment, treat it as code
            code = Code(
                image=seg_data,
                format="blob",
                arch=self.arch,
                mode=self.mode,
                base=seg_addr,
                entry=self.entry,
                exits=self.exits,
            )
            self.code_segments.append(code)
        else:
            # Otherwise, treat it as Memory
            data = Memory(seg_addr, seg_size)
            data.set(seg_data)
            self.data_segments.append(data)

    def apply(self, emulator: Emulator, override: bool = True):
        for code in self.code_segments:
            code.apply(emulator)
        for data in self.data_segments:
            data.apply(emulator)
